"""Open Response Assessment events model event fields definitions."""

import sys
from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import UUID

from pydantic import constr

from ralph.models.edx.base import AbstractBaseEventField, BaseModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class ORAGetPeerSubmissionEventField(AbstractBaseEventField):
    """Pydantic model for `openassessmentblock.get_peer_submission`.`event` field.

    Attributes:
        course_id (str): Consists of the course identifier including the assessment.
        item_id (str): Consists of the locator string that identifies the problem in
            the course.
        requesting_student_id (str): Consists of the course-specific anonymized user ID
            of the learner who retrieved the response for peer assessment.
        submission_returned_uuid (str): Consists of the unique identifier of the
            response that was retrieved for assessment. Set to `None` if no assessment
            available.
    """

    course_id: constr(max_length=255)
    item_id: constr(
        regex=(r"^block-v1:.+\+.+\+.+type@openassessment+block@[a-f0-9]{32}$")
    )
    requesting_student_id: str
    submission_returned_uuid: Union[str, None]


class ORAGetSubmissionForStaffGradingEventField(AbstractBaseEventField):
    """Pydantic model for `openassessmentblock.get_submission_for_staff_grading`.
    `event` field.

    Attributes:
        item_id (str): Consists of the locator string that identifies the problem in
            the course.
        submission_returned_uuid (str): Consists of the unique identifier of the
            response that was retrieved for assessment. Set to `None` if no assessment
            available.
        requesting_staff_id (str): Consists of the course-specific anonymized user ID
            of the course team member who is retrieved the response for grading.
        type (str): Consists of the type of staff grading that is being performed.
            Currently, set to `full-grade`.
    """  # noqa: D205

    item_id: constr(
        regex=(r"^block-v1:.+\+.+\+.+type@openassessment+block@[a-f0-9]{32}$")
    )
    submission_returned_uuid: Union[str, None]
    requesting_staff_id: str
    type: Literal["full-grade"]


class ORAAssessEventPartsCriterionField(BaseModelWithConfig):
    """Pydantic model for assessment `event`.`parts`.`criterion` field.

    Attributes:
        name (str): Consists of the criterion name.
        points_possible (int): Consists of the maximum number of points
            allocated to the criterion.
    """

    name: str
    points_possible: int


class ORAAssessEventPartsField(BaseModelWithConfig):
    """Pydantic model for assessment `event`.`parts` field.

    Attributes:
        option (str): Consists of the option that the learner selected for it.
        criterion (dict): see ORAAssessEventPartsCriterionField.
        feedback (str): Consists of feedback comments that the learner could have
            supplied.
    """

    option: str
    criterion: ORAAssessEventPartsCriterionField
    feedback: Optional[str]


class ORAAssessEventRubricField(BaseModelWithConfig):
    """Pydantic model for assessment `event`.`rubric` field.

    This field is defined in:
    - `openassessmentblock.peer_assess`
    - `openassessmentblock.self_assess`
    - `openassessmentblock.staff_assess`

    Attributes:
        content_hash: Consists of the identifier of the rubric that the learner used to
            assess the response.
    """

    content_hash: constr(regex=r"^[a-f0-9]{1,40}$")


class ORAAssessEventField(AbstractBaseEventField):
    """Pydantic model for assessment `event` field.

    This field is defined in:
        - `openassessmentblock.peer_assess`
        - `openassessmentblock.self_assess`
        - `openassessmentblock.staff_assess`

    Attributes:
        feedback (str): Consists of the learner's comments about the submitted response.
        parts (list): see ORAAssessEventPartsField.
        rubric (dict): see ORAPeerAssessEventRubricField.
        scored_at (datetime): Consists of the timestamp for when the assessment was
            submitted.
        scorer_id (str): Consists of the course-specific anonymized user ID of the
            learner who submitted the assessment.
        score_type (str): Consists of either `PE` value for a peer assessment, `SE` for
            a self assessment or `ST` for a staff assessment.
        submission_uuid (str): Consists of the unique identifier for the submitted
            response.
    """

    feedback: str
    parts: List[ORAAssessEventPartsField]
    rubric: ORAAssessEventRubricField
    scored_at: datetime
    scorer_id: constr(max_length=40)
    score_type: Literal["PE", "SE", "ST"]
    submission_uuid: UUID


class ORAStaffAssessEventField(ORAAssessEventField):
    """Pydantic model for `openassessmentblock.staff_assess`.`event` field.

    Attributes:
        type (str): Consists of the type of staff grading that is being performed. Can
            be either equal to `regrade` in the case of a grade override or
            `full-grade` in the case of an included staff assessment step.
    """

    type: Literal["regrade", "full-grade"]


class ORASubmitFeedbackOnAssessmentsEventField(AbstractBaseEventField):
    """Pydantic model for `openassessmentblock.submit_feedback_on_assessments`.
    `event` field.

    Attributes:
        feedback_text (str): Consists of the learner's comments about the assessment
            process.
        options (list): Consists of the label of each checkbox option that the learner
            selected to evaluate the assessment process.
        submission_uuid (str): Consists of the unique identifier for the feedback.
    """  # noqa: D205

    feedback_text: str
    options: List[str]
    submission_uuid: UUID


class ORACreateSubmissionEventAnswerField(BaseModelWithConfig):
    """Pydantic model for `openassessmentblock.create_submission`.`event`.`answer`
    field.

    Attributes:
        parts (dict): Consists of a key-value dictionary with all answers text.
        file_keys (list): Consists of a list of file identifiers if files are given for
            answer.
        files_description (list): Consists of a list of file descriptions if files are
            given for answer.
    """  # noqa: D205

    parts: List[Dict[Literal["text"], str]]
    file_keys: Optional[List[str]]
    files_descriptions: Optional[List[str]]


class ORACreateSubmissionEventField(AbstractBaseEventField):
    """Pydantic model for `openassessmentblock.create_submission`.`event` field.

    Attributes:
        answer (dict): see ORACreateSubmissionEventAnswerField.
        attempt_number (int): Consists of the number of submission attempts. Currently,
            this value is set to 1.
        created_at (datetime): Consists of the timestamp for when the learner submitted
            the response.
        submitted_at (datetime): Consists of the timestamp for when the learner
            submitted the response. This value is the same as `submitted_at`.
        submission_uuid (str): Consists of the unique identifier of the response.
    """

    answer: ORACreateSubmissionEventAnswerField
    attempt_number: int
    created_at: datetime
    submitted_at: datetime
    submission_uuid: UUID


class ORASaveSubmissionEventSavedResponseField(BaseModelWithConfig):
    """Pydantic model for `openassessmentblock.save_submission`.`saved_response` field.

    Attributes:
        text (str): Consists of the response text.
        file_upload_key (str): Consists of the AWS S3 key that identifies the location
            of the uploaded file on the Amazon S3 storage service. Only present when
            responses include an image, .pdf, or other file.
    """

    text: str
    file_upload_key: Optional[str]


class ORASaveSubmissionEventField(AbstractBaseEventField):
    """Pydantic model for `openassessmentblock.save_submission`.`event` field.

    Attributes:
        saved_response (str): Consists of a JSON string of the users saved responses.
            Note::
                Responses have a length limit of 100000 in the front-end but not in the
                back-end. Events are truncated at `TRACK_MAX_EVENT` which is 50000 by
                default. Also, the `eventtracking.backends.logger.LoggerBackend`
                silently drops events when they exceed `TRACK_MAX_EVENT`.
    """

    saved_response: ORASaveSubmissionEventSavedResponseField


class ORAStudentTrainingAssessExampleEventField(AbstractBaseEventField):
    """Pydantic model for `openassessment.student_training_assess_example`.`event`
    field.

    Attributes:
        corrections (dict): Consists of a set of name/value pairs that identify
            criteria for which the learner selected a different option than the course
            team.
        options_selected (dict): Consists of a set of name/value pairs that identify
            the option that the learner selected for each criterion in the rubric.
        submission_uuid (str): Consists of the unique identifier of the response.
    """  # noqa: D205

    corrections: Dict[str, str]
    options_selected: Dict[str, str]
    submission_uuid: UUID


class ORAUploadFileEventField(BaseModelWithConfig):
    """Pydantic model for `openassessment.upload_file`.`event` field.

    Attributes:
        fileName (str): Consists of the name of the uploaded file.
        fileSize (int): Consists of the bytes size of the uploaded file.
        fileType (str): Consists of the MIME type of the uploaded file.
    """

    fileName: constr(max_length=255)
    fileSize: int
    fileType: str



"""Virtual classroom xAPI event definitions."""

from datetime import datetime

from ...selector import selector
from ..base.statements import BaseXapiStatement
from ..concepts.activity_types.acrossx_profile import MessageActivity
from ..concepts.activity_types.scorm_profile import CMIInteractionActivity
from ..concepts.activity_types.virtual_classroom import VirtualClassroomActivity
from ..concepts.verbs.acrossx_profile import PostedVerb
from ..concepts.verbs.activity_streams_vocabulary import JoinVerb, LeaveVerb
from ..concepts.verbs.adl_vocabulary import AnsweredVerb, AskedVerb
from ..concepts.verbs.scorm_profile import InitializedVerb, TerminatedVerb
from ..concepts.verbs.virtual_classroom import (
    LoweredHandVerb,
    MutedVerb,
    RaisedHandVerb,
    SharedScreenVerb,
    StartedCameraVerb,
    StoppedCameraVerb,
    UnmutedVerb,
    UnsharedScreenVerb,
)
from .contexts import (
    VirtualClassroomAnsweredPollContext,
    VirtualClassroomContext,
    VirtualClassroomInitializedContext,
    VirtualClassroomJoinedContext,
    VirtualClassroomPostedPublicMessageContext,
    VirtualClassroomStartedPollContext,
    VirtualClassroomTerminatedContext,
)
from .results import VirtualClassroomAnsweredPollResult


class BaseVirtualClassroomStatement(BaseXapiStatement):
    """Pydantic model for Virtual Classroom core statements.

    Attributes:
        timestamp (datetime): Consists of the timestamp of when the event occurred.
    """

    timestamp: datetime


# Mandatory statements


class VirtualClassroomInitialized(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom initialized statement.

    Example: John has entered the virtual classroom as the first user.

    Attributes:
        verb (dict): See InitializedVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomInitializedContext.
    """

    __selector__ = selector(
        verb__id="http://adlnet.gov/expapi/verbs/initialized",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: InitializedVerb = InitializedVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomInitializedContext


class VirtualClassroomJoined(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom joined statement.

    Example: Jane has entered the virtual classroom.

    Attributes:
        verb (dict): See JoinVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomJoinedContext.
    """

    __selector__ = selector(
        verb__id="http://activitystrea.ms/join",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: JoinVerb = JoinVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomJoinedContext


class VirtualClassroomLeft(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom joined statement.

    Example: Jane has left the virtual classroom.

    Attributes:
        verb (dict): See LeaveVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomContext.

    """

    __selector__ = selector(
        verb__id="http://activitystrea.ms/leave",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: LeaveVerb = LeaveVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomTerminated(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom terminated statement.

    Example: Jane has terminated the virtual classroom.

    Attributes:
        verb (dict): See TerminatedVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomTerminatedContext.

    """

    __selector__ = selector(
        verb__id="http://adlnet.gov/expapi/verbs/terminated",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: TerminatedVerb = TerminatedVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomTerminatedContext


# Recommended statements


class VirtualClassroomMuted(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom muted statement.

    Example: Jane has muted herself or has been muted.

    Attributes:
        verb (dict): See MutedVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomMutedContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/muted",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: MutedVerb = MutedVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomUnmuted(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom unmuted statement.

    Example: Jane has unmuted herself or has been unmuted.

    Attributes:
        verb (dict): See UnmutedVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/unmuted",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: UnmutedVerb = UnmutedVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomStartedCamera(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom started camera statement.

    Example: Jane has started her camera.

    Attributes:
        verb (dict): See StartedCameraVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/started-camera",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: StartedCameraVerb = StartedCameraVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomStoppedCamera(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom stopped camera statement.

    Example: Jane has stopped her camera.

    Attributes:
        verb (dict): See StoppedCameraVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomStoppedCameraContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/stopped-camera",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: StoppedCameraVerb = StoppedCameraVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomSharedScreen(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom shared screen statement.

    Example: Jane has shared her screen.

    Attributes:
        verb (dict): See SharedScreenVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomSharedScreenContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/shared-screen",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: SharedScreenVerb = SharedScreenVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomUnsharedScreen(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom unshared screen statement.

    Example: Jane has unshared her screen or her screen has been unshared by the
    moderator.

    Attributes:
        verb (dict): See UnsharedScreenVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomUnsharedScreenContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/unshared-screen",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: UnsharedScreenVerb = UnsharedScreenVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomRaisedHand(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom raised hand statement.

    Example: Jane has raised the hand.

    Attributes:
        verb (dict): See RaisedHandVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomRaisedHandContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/raised-hand",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: RaisedHandVerb = RaisedHandVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomLoweredHand(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom lowered hand statement.

    Example: Jane has lowered the hand.

    Attributes:
        verb (dict): See LoweredHandVerb.
        object (dict): See VirtualClassroomActivity.
        context (dict): See VirtualClassroomLoweredHandContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/virtual-classroom/verbs/lowered-hand",
        object__definition__type=(
            "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
        ),
    )

    verb: LoweredHandVerb = LoweredHandVerb()
    object: VirtualClassroomActivity
    context: VirtualClassroomContext


class VirtualClassroomStartedPoll(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom started poll statement.

    Example: A poll has been started in the virtual classroom in order
    to collect participants opinions about a given question.

    Attributes:
        verb (dict): See AskedVerb.
        object (dict): See CMIInteractionActivity.
        context (dict): See VirtualClassroomLoweredHandContext.

    """

    __selector__ = selector(
        verb__id="http://adlnet.gov/expapi/verbs/asked",
        object__definition__type=(
            "http://adlnet.gov/expapi/activities/cmi.interaction"
        ),
    )

    verb: AskedVerb = AskedVerb()
    object: CMIInteractionActivity
    context: VirtualClassroomStartedPollContext


class VirtualClassroomAnsweredPoll(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom answered poll statement.

    Example: John answered to a poll question.

    Attributes:
        verb (dict): See AskedVerb.
        object (dict): See CMIInteractionActivity.
        context (dict): See VirtualClassroomAnsweredPollContext.
        result (dict): See AnsweredPollResult.
        result (dict): See AnsweredPollResult.
        timestamp (datetime): Consists of the timestamp of when the event occurred.
        result (dict): See AnsweredPollResult.
        timestamp (datetime): Consists of the timestamp of when the event occurred.
    """

    __selector__ = selector(
        verb__id="http://adlnet.gov/expapi/verbs/answered",
        object__definition__type=(
            "http://adlnet.gov/expapi/activities/cmi.interaction"
        ),
    )

    verb: AnsweredVerb = AnsweredVerb()
    object: CMIInteractionActivity
    context: VirtualClassroomAnsweredPollContext
    result: VirtualClassroomAnsweredPollResult


class VirtualClassroomPostedPublicMessage(BaseVirtualClassroomStatement):
    """Pydantic model for virtual classroom posted public message statement.

    Example: John sent a message in the public chat.

    Attributes:
        verb (dict): See PostedVerb.
        object (dict): See MessageActivity.
        context (dict): See VirtualClassroomPostedPublicMessageContext.

    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/acrossx/verbs/posted",
        object__definition__type="https://w3id.org/xapi/acrossx/activities/message",
    )

    verb: PostedVerb = PostedVerb()
    object: MessageActivity
    context: VirtualClassroomPostedPublicMessageContext



"""Main module for Ralph's LRS API."""
from functools import lru_cache
from typing import Any, Dict, List, Union
from urllib.parse import urlparse

import sentry_sdk
from fastapi import Depends, FastAPI

from ralph.conf import settings

from .. import __version__
from .auth import get_authenticated_user
from .auth.user import AuthenticatedUser
from .routers import health, statements


@lru_cache(maxsize=None)
def get_health_check_routes() -> List:
    """Return the health check routes."""
    return [route.path for route in health.router.routes]


def filter_transactions(event: Dict, hint) -> Union[Dict, None]:  # noqa: ARG001
    """Filter transactions for Sentry."""
    url = urlparse(event["request"]["url"])

    if settings.SENTRY_IGNORE_HEALTH_CHECKS and url.path in get_health_check_routes():
        return None

    return event


if settings.SENTRY_DSN is not None:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        traces_sample_rate=settings.SENTRY_LRS_TRACES_SAMPLE_RATE,
        release=__version__,
        environment=settings.EXECUTION_ENVIRONMENT,
        max_breadcrumbs=50,
        before_send_transaction=filter_transactions,
    )

app = FastAPI()
app.include_router(statements.router)
app.include_router(health.router)


@app.get("/whoami")
async def whoami(
    user: AuthenticatedUser = Depends(get_authenticated_user),
) -> Dict[str, Any]:
    """Return the current user's username along with their scopes."""
    return {"agent": user.agent, "scopes": user.scopes}



"""API routes related to statements."""

import json
import logging
from datetime import datetime
from typing import Dict, List, Literal, Optional, Union
from urllib.parse import ParseResult, urlencode
from uuid import UUID, uuid4

from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    HTTPException,
    Query,
    Request,
    Response,
    Security,
    status,
)
from fastapi.dependencies.models import Dependant
from pydantic import parse_obj_as
from pydantic.types import Json
from typing_extensions import Annotated

from ralph.api.auth import get_authenticated_user
from ralph.api.auth.user import AuthenticatedUser
from ralph.api.forwarding import forward_xapi_statements, get_active_xapi_forwardings
from ralph.api.models import ErrorDetail, LaxStatement
from ralph.backends.loader import get_lrs_backends
from ralph.backends.lrs.base import (
    AgentParameters,
    BaseAsyncLRSBackend,
    BaseLRSBackend,
    RalphStatementsQuery,
)
from ralph.conf import settings
from ralph.exceptions import BackendException, BadFormatException
from ralph.models.xapi.base.agents import (
    BaseXapiAgent,
    BaseXapiAgentWithAccount,
    BaseXapiAgentWithMbox,
    BaseXapiAgentWithMboxSha1Sum,
    BaseXapiAgentWithOpenId,
)
from ralph.models.xapi.base.common import IRI
from ralph.utils import (
    await_if_coroutine,
    get_backend_class,
    now,
    statements_are_equivalent,
)

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/xAPI/statements",
    dependencies=[Depends(get_authenticated_user)],
)


BACKEND_CLIENT: Union[BaseLRSBackend, BaseAsyncLRSBackend] = get_backend_class(
    backends=get_lrs_backends(), name=settings.RUNSERVER_BACKEND
)()

POST_PUT_RESPONSES = {
    400: {
        "model": ErrorDetail,
        "description": "The request was invalid.",
    },
    409: {
        "model": ErrorDetail,
        "description": "Statements had a conflict with existing statements.",
    },
}


def _enrich_statement_with_id(statement: dict) -> None:
    # id: Statement UUID identifier.
    # https://github.com/adlnet/xAPI-Spec/blob/master/xAPI-Data.md#24-statement-properties
    statement["id"] = str(statement.get("id", uuid4()))


def _enrich_statement_with_stored(statement: dict) -> None:
    # stored: The time at which a Statement is stored by the LRS.
    # https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Data.md#248-stored
    statement["stored"] = now()


def _enrich_statement_with_timestamp(statement: dict) -> None:
    # timestamp: Time of the action. If not provided, it takes the same value as stored.
    # https://github.com/adlnet/xAPI-Spec/blob/master/xAPI-Data.md#247-timestamp
    statement["timestamp"] = statement.get("timestamp", statement["stored"])


def _enrich_statement_with_authority(
    statement: dict, current_user: AuthenticatedUser
) -> None:
    # authority: Information about whom or what has asserted the statement is true.
    # https://github.com/adlnet/xAPI-Spec/blob/master/xAPI-Data.md#249-authority
    statement["authority"] = current_user.agent


def _parse_agent_parameters(agent_obj: dict) -> AgentParameters:
    """Parse a dict and return an AgentParameters object to use in queries."""
    # Transform agent to `dict` as FastAPI cannot parse JSON (seen as string)

    agent = parse_obj_as(BaseXapiAgent, agent_obj)

    agent_query_params = {}
    if isinstance(agent, BaseXapiAgentWithMbox):
        agent_query_params["mbox"] = agent.mbox
    elif isinstance(agent, BaseXapiAgentWithMboxSha1Sum):
        agent_query_params["mbox_sha1sum"] = agent.mbox_sha1sum
    elif isinstance(agent, BaseXapiAgentWithOpenId):
        agent_query_params["openid"] = agent.openid
    elif isinstance(agent, BaseXapiAgentWithAccount):
        agent_query_params["account__name"] = agent.account.name
        agent_query_params["account__home_page"] = agent.account.homePage

    # Overwrite `agent` field
    return AgentParameters.construct(**agent_query_params)


def strict_query_params(request: Request) -> None:
    """Raise a 400 error when using extra query parameters."""
    dependant: Dependant = request.scope["route"].dependant
    allowed_params = [
        param.alias
        for dependency in dependant.dependencies
        for param in dependency.query_params
    ]
    allowed_params += [modelfield.alias for modelfield in dependant.query_params]
    for requested_param in request.query_params.keys():
        if requested_param not in allowed_params:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"The following parameter is not allowed: `{requested_param}`",
            )


@router.get("")
@router.get("/")
async def get(  # noqa: PLR0913
    request: Request,
    current_user: Annotated[
        AuthenticatedUser,
        Security(get_authenticated_user, scopes=["statements/read/mine"]),
    ],
    ###
    # Query string parameters defined by the LRS specification
    ###
    statement_id: Optional[str] = Query(
        None, description="Id of Statement to fetch", alias="statementId"
    ),
    voided_statement_id: Optional[str] = Query(
        None,
        description="**Not implemented** Id of voided Statement to fetch",
        alias="voidedStatementId",
    ),
    agent: Optional[Json] = Query(
        None,
        description=(
            "Filter, only return Statements for which the specified "
            "Agent or Group is the Actor or Object of the Statement"
        ),
    ),
    verb: Optional[str] = Query(
        None,
        description="Filter, only return Statements matching the specified Verb id",
    ),
    activity: Optional[IRI] = Query(
        None,
        description=(
            "Filter, only return Statements for which the Object "
            "of the Statement is an Activity with the specified id"
        ),
    ),
    registration: Optional[UUID] = Query(
        None,
        description=(
            "**Not implemented** "
            "Filter, only return Statements matching the specified registration id"
        ),
    ),
    related_activities: Optional[bool] = Query(
        False,
        description=(
            "**Not implemented** "
            "Apply the Activity filter broadly. Include Statements for which "
            "the Object, any of the context Activities, or any of those properties "
            "in a contained SubStatement match the Activity parameter, "
            "instead of that parameter's normal behaviour"
        ),
    ),
    related_agents: Optional[bool] = Query(
        False,
        description=(
            "**Not implemented** "
            "Apply the Agent filter broadly. Include Statements for which "
            "the Actor, Object, Authority, Instructor, Team, or any of these "
            "properties in a contained SubStatement match the Agent parameter, "
            "instead of that parameter's normal behaviour."
        ),
    ),
    since: Optional[datetime] = Query(
        None,
        description=(
            "Only Statements stored since the "
            "specified Timestamp (exclusive) are returned"
        ),
    ),
    until: Optional[datetime] = Query(
        None,
        description=(
            "Only Statements stored at or "
            "before the specified Timestamp are returned"
        ),
    ),
    limit: Optional[int] = Query(
        settings.RUNSERVER_MAX_SEARCH_HITS_COUNT,
        ge=0,
        description=(
            "Maximum number of Statements to return. "
            "0 indicates return the maximum the server will allow"
        ),
    ),
    format: Optional[Literal["ids", "exact", "canonical"]] = Query(  # noqa: ARG001
        "exact",
        description=(
            "**Not implemented** "
            'If "ids", only include minimum information necessary in Agent, Activity, '
            "Verb and Group Objects to identify them. For Anonymous Groups this means "
            "including the minimum information needed to identify each member. "
            'If "exact", return Agent, Activity, Verb and Group Objects populated '
            "exactly as they were when the Statement was received. An LRS requesting "
            "Statements for the purpose of importing them would use a format of "
            '"exact" in order to maintain Statement Immutability. '
            'If "canonical", return Activity Objects and Verbs populated with the '
            "canonical definition of the Activity Objects and Display of the Verbs "
            "as determined by the LRS, after applying the language filtering process "
            "defined below, and return the original Agent and Group Objects "
            'as in "exact" mode.'
        ),
    ),
    attachments: Optional[bool] = Query(  # noqa: ARG001
        False,
        description=(
            "**Not implemented** "
            'If "true", the LRS uses the multipart response format and includes '
            'all attachments as described previously. If "false", the LRS sends '
            "the prescribed response with Content-Type application/json and "
            "does not send attachment data."
        ),
    ),
    ascending: Optional[bool] = Query(  # noqa: ARG001
        False, description='If "true", return results in ascending order of stored time'
    ),
    mine: Optional[bool] = Query(
        False,
        description=(
            'If "true", return only the results for which the authority matches the '
            '"agent" associated to the user that is making the query.'
        ),
    ),
    ###
    # Private use query string parameters
    ###
    search_after: Optional[str] = Query(  # noqa: ARG001
        None,
        description=(
            "Sorting data to allow pagination through large number of search results. "
            "NB: for internal use, not part of the LRS specification."
        ),
    ),
    pit_id: Optional[str] = Query(  # noqa: ARG001
        None,
        description=(
            "Point-in-time ID to ensure consistency of search requests through "
            "multiple pages."
            "NB: for internal use, not part of the LRS specification."
        ),
    ),
    _=Depends(strict_query_params),
) -> Dict:
    """Read a single xAPI Statement or multiple xAPI Statements.

    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#213-get-statements
    """
    # Make sure the limit does not go above max from settings
    limit = min(limit, settings.RUNSERVER_MAX_SEARCH_HITS_COUNT)

    # 400 Bad Request for requests using both `statement_id` and `voided_statement_id`
    if (statement_id is not None) and (voided_statement_id is not None):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                "Query parameters cannot include both `statementId`"
                "and `voidedStatementId`"
            ),
        )

    # 400 Bad Request for any request containing `statement_id` or
    # `voided_statement_id` and any other parameter besides `attachments` or `format`.
    # NB: `limit` and `ascending` are not handled to simplify implementation, and as it
    # has no incidence on UX, when fetching a single statement
    excluded_params = [
        agent,
        verb,
        activity,
        registration,
        related_activities,
        related_agents,
        since,
        until,
    ]
    # NB: bool(param) relies on all defaults being None, 0, or False
    if (statement_id or voided_statement_id) and any(
        bool(param) for param in excluded_params
    ):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                "Querying by id only accepts `attachments` and `format` as extra "
                "parameters"
            ),
        )

    query_params = dict(request.query_params)

    # Parse the "agent" parameter (JSON) into multiple string parameters
    if query_params.get("agent") is not None:
        # Overwrite `agent` field
        query_params["agent"] = _parse_agent_parameters(
            json.loads(query_params["agent"])
        )

    # mine: If using scopes, only restrict users with limited scopes
    if settings.LRS_RESTRICT_BY_SCOPES:
        if not current_user.scopes.is_authorized("statements/read"):
            mine = True
    # mine: If using only authority, always restrict (otherwise, use the default value)
    elif settings.LRS_RESTRICT_BY_AUTHORITY:
        mine = True

    # Filter by authority if using `mine`
    if mine:
        query_params["authority"] = _parse_agent_parameters(current_user.agent)

    if "mine" in query_params:
        query_params.pop("mine")

    # Query Database
    try:
        query_result = await await_if_coroutine(
            BACKEND_CLIENT.query_statements(
                params=RalphStatementsQuery.construct(
                    **{**query_params, "limit": limit}
                ),
                target=current_user.target,
            )
        )
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    # Prepare the link to get the next page of the request, while preserving the
    # consistency of search results.
    # NB: There is an unhandled edge case where the total number of results is
    # exactly a multiple of the "limit", in which case we'll offer an extra page
    # with 0 results.
    response = {}
    if len(query_result.statements) == limit:
        # Search after relies on sorting info located in the last hit
        path = request.url.path
        query = dict(request.query_params)

        query.update(
            {
                "pit_id": query_result.pit_id,
                "search_after": query_result.search_after,
            }
        )

        response.update(
            {
                "more": ParseResult(
                    scheme="",
                    netloc="",
                    path=path,
                    params="",
                    query=urlencode(query),
                    fragment="",
                ).geturl(),
            }
        )

    return {**response, "statements": query_result.statements}


@router.put("/", responses=POST_PUT_RESPONSES, status_code=status.HTTP_204_NO_CONTENT)
@router.put("", responses=POST_PUT_RESPONSES, status_code=status.HTTP_204_NO_CONTENT)
async def put(
    current_user: Annotated[
        AuthenticatedUser,
        Security(get_authenticated_user, scopes=["statements/write"]),
    ],
    statement: LaxStatement,
    background_tasks: BackgroundTasks,
    statement_id: UUID = Query(alias="statementId"),
    _=Depends(strict_query_params),
) -> None:
    """Store a single statement as a single member of a set.

    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#211-put-statements
    """
    statement_as_dict = statement.dict(exclude_unset=True)
    statement_id = str(statement_id)

    statement_as_dict.update(id=str(statement_as_dict.get("id", statement_id)))
    if statement_id != statement_as_dict["id"]:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="xAPI statement id does not match given statementId",
        )

    # Enrich statement before forwarding (NB: id is already set)
    _enrich_statement_with_stored(statement_as_dict)
    _enrich_statement_with_timestamp(statement_as_dict)

    if get_active_xapi_forwardings():
        background_tasks.add_task(
            forward_xapi_statements, statement_as_dict, method="put"
        )

    # Finish enriching statements after forwarding
    _enrich_statement_with_authority(statement_as_dict, current_user)

    try:
        if isinstance(BACKEND_CLIENT, BaseLRSBackend):
            existing_statements = list(
                BACKEND_CLIENT.query_statements_by_ids(
                    ids=[statement_id], target=current_user.target
                )
            )
        else:
            existing_statements = [
                x
                async for x in BACKEND_CLIENT.query_statements_by_ids(
                    ids=[statement_id], target=current_user.target
                )
            ]
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    if existing_statements:
        # The LRS specification calls for deep comparison of duplicate statement ids.
        # In the case that the current statement is not equivalent to one found
        # in the database we return a 409, otherwise the usual 204.
        for existing in existing_statements:
            if not statements_are_equivalent(statement_as_dict, existing):
                raise HTTPException(
                    status_code=status.HTTP_409_CONFLICT,
                    detail="A different statement already exists with the same ID",
                )
        return

    # For valid requests, perform the bulk indexing of all incoming statements
    try:
        success_count = await await_if_coroutine(
            BACKEND_CLIENT.write(
                data=[statement_as_dict],
                target=current_user.target,
                ignore_errors=False,
            )
        )
    except (BackendException, BadFormatException) as exc:
        logger.error("Failed to index submitted statement")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Statement indexation failed",
        ) from exc

    logger.info("Indexed %d statements with success", success_count)


@router.post("/", responses=POST_PUT_RESPONSES)
@router.post("", responses=POST_PUT_RESPONSES)
async def post(  # noqa: PLR0912
    current_user: Annotated[
        AuthenticatedUser,
        Security(get_authenticated_user, scopes=["statements/write"]),
    ],
    statements: Union[LaxStatement, List[LaxStatement]],
    background_tasks: BackgroundTasks,
    response: Response,
    _=Depends(strict_query_params),
) -> Union[List, None]:
    """Store a set of statements (or a single statement as a single member of a set).

    NB: at this time, using POST to make a GET request, is not supported.
    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#212-post-statements
    """
    # As we accept both a single statement as a dict, and multiple statements as a list,
    # we need to normalize the data into a list in all cases before we can process it.
    if not isinstance(statements, list):
        statements = [statements]

    # Enrich statements before forwarding
    statements_dict = {}
    for statement in (x.dict(exclude_unset=True) for x in statements):
        _enrich_statement_with_id(statement)
        # Requests with duplicate statement IDs are considered invalid
        if statement["id"] in statements_dict:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Duplicate statement IDs in the list of statements",
            )
        _enrich_statement_with_stored(statement)
        _enrich_statement_with_timestamp(statement)
        _enrich_statement_with_authority(statement, current_user)
        statements_dict[statement["id"]] = statement

    # Forward statements
    if get_active_xapi_forwardings():
        background_tasks.add_task(
            forward_xapi_statements, list(statements_dict.values()), method="post"
        )

    try:
        if isinstance(BACKEND_CLIENT, BaseLRSBackend):
            existing_statements = list(
                BACKEND_CLIENT.query_statements_by_ids(
                    ids=list(statements_dict), target=current_user.target
                )
            )
        else:
            existing_statements = [
                x
                async for x in BACKEND_CLIENT.query_statements_by_ids(
                    ids=list(statements_dict), target=current_user.target
                )
            ]
    except BackendException as error:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="xAPI statements query failed",
        ) from error

    # If there are duplicate statements, remove them from our id list and
    # dictionary for insertion. We will return the shortened list of ids below
    # so that consumers can derive which statements were inserted and which
    # were skipped for being duplicates.
    # See: https://github.com/openfun/ralph/issues/345
    if existing_statements:
        existing_ids = set()
        for existing in existing_statements:
            existing_ids.add(existing["id"])

            # The LRS specification calls for deep comparison of duplicates. This
            # is done here. If they are not exactly the same, we raise an error.
            if not statements_are_equivalent(statements_dict[existing["id"]], existing):
                raise HTTPException(
                    status_code=status.HTTP_409_CONFLICT,
                    detail="Differing statements already exist with the same ID: "
                    f"{existing['id']}",
                )

        # Filter existing statements from the incoming statements
        statements_dict = {
            key: value
            for key, value in statements_dict.items()
            if key not in existing_ids
        }
        if not statements_dict:
            response.status_code = status.HTTP_204_NO_CONTENT
            return

    # For valid requests, perform the bulk indexing of all incoming statements
    try:
        success_count = await await_if_coroutine(
            BACKEND_CLIENT.write(
                data=statements_dict.values(),
                target=current_user.target,
                ignore_errors=False,
            )
        )
    except (BackendException, BadFormatException) as exc:
        logger.error("Failed to index submitted statements")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Statements bulk indexation failed",
        ) from exc

    logger.info("Indexed %d statements with success", success_count)

    # Return the list of IDs in the same order they were stored
    return list(statements_dict)



"""Video xAPI events context fields definitions."""

import sys
from typing import List, Optional, Union
from uuid import UUID

from pydantic import Field, NonNegativeFloat, validator

from ..base.contexts import BaseXapiContext, BaseXapiContextContextActivities
from ..base.unnested_objects import BaseXapiActivity
from ..concepts.activity_types.scorm_profile import ProfileActivity
from ..concepts.constants.video import (
    CONTEXT_EXTENSION_CC_ENABLED,
    CONTEXT_EXTENSION_CC_SUBTITLE_LANG,
    CONTEXT_EXTENSION_COMPLETION_THRESHOLD,
    CONTEXT_EXTENSION_FULL_SCREEN,
    CONTEXT_EXTENSION_LENGTH,
    CONTEXT_EXTENSION_SCREEN_SIZE,
    CONTEXT_EXTENSION_SESSION_ID,
    CONTEXT_EXTENSION_SPEED,
    CONTEXT_EXTENSION_USER_AGENT,
    CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE,
    CONTEXT_EXTENSION_VOLUME,
)
from ..config import BaseExtensionModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class VideoProfileActivity(ProfileActivity):
    """Pydantic model for video profile `Activity` type.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/video`.
    """

    id: Literal["https://w3id.org/xapi/video"] = "https://w3id.org/xapi/video"


class VideoContextContextActivities(BaseXapiContextContextActivities):
    """Pydantic model for video `context`.`contextActivities` property.

    Attributes:
        category (dict or list): see VideoProfileActivity.
    """

    category: Union[
        VideoProfileActivity, List[Union[VideoProfileActivity, BaseXapiActivity]]
    ]

    @validator("category")
    @classmethod
    def check_presence_of_profile_activity_category(
        cls,
        value: Union[
            VideoProfileActivity, List[Union[VideoProfileActivity, BaseXapiActivity]]
        ],
    ) -> Union[
        VideoProfileActivity, List[Union[VideoProfileActivity, BaseXapiActivity]]
    ]:
        """Check that the category list contains a `VideoProfileActivity`."""
        if isinstance(value, VideoProfileActivity):
            return value
        for activity in value:
            if isinstance(activity, VideoProfileActivity):
                return value
        raise ValueError(
            "The `context.contextActivities.category` field should contain at least "
            "one valid `VideoProfileActivity`"
        )


class BaseVideoContext(BaseXapiContext):
    """Pydantic model for video core `context` property.

    Attributes:
        contextActivities (dict): see VideoContextContextActivities.
    """

    contextActivities: VideoContextContextActivities


class VideoContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for video core context `extensions` property.

    Attributes:
        session (uuid): Consists of the ID of the active session.
    """

    session_id: Optional[UUID] = Field(alias=CONTEXT_EXTENSION_SESSION_ID)


class VideoInitializedContextExtensions(VideoContextExtensions):
    """Pydantic model for video initialized `context` `extensions` property.

    Attributes:
        length (float): Consists of the length of the video.
        ccSubtitleEnabled (bool): Indicates whether subtitle or closed captioning is
            enabled.
        ccSubtitleLanguage (str): Consists of the language of subtitle or closed
            captioning.
        fullScreen (bool): Indicates whether the video is played in full screen mode.
        screenSize (str): Consists of the device playback screen size or the maximum
            available screen size for Video playback.
        videoPlaybackSize (str): Consists of the size in Width x Height of the video as
            viewed by the user.
        speed (str): Consists of the play back speed.
        userAgent (str): Consists of the User Agent string of the browser,
            if the video is launched in browser.
        volume (int): Consists of the volume of the video.
        completionThreshold (float): Consists of the percentage of media that should be
            consumed to trigger a completion.
    """

    length: NonNegativeFloat = Field(alias=CONTEXT_EXTENSION_LENGTH)
    ccSubtitleEnabled: Optional[bool] = Field(alias=CONTEXT_EXTENSION_CC_ENABLED)
    ccSubtitleLang: Optional[str] = Field(alias=CONTEXT_EXTENSION_CC_SUBTITLE_LANG)
    fullScreen: Optional[bool] = Field(alias=CONTEXT_EXTENSION_FULL_SCREEN)
    screenSize: Optional[str] = Field(alias=CONTEXT_EXTENSION_SCREEN_SIZE)
    videoPlaybackSize: Optional[str] = Field(
        alias=CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE
    )
    speed: Optional[str] = Field(alias=CONTEXT_EXTENSION_SPEED)
    userAgent: Optional[str] = Field(alias=CONTEXT_EXTENSION_USER_AGENT)
    volume: Optional[int] = Field(alias=CONTEXT_EXTENSION_VOLUME)
    completionThreshold: Optional[float] = Field(
        alias=CONTEXT_EXTENSION_COMPLETION_THRESHOLD
    )


class VideoBrowsingContextExtensions(VideoContextExtensions):
    """Pydantic model for video browsing `context`.`extensions` property.

    Such field is used in `paused`, `completed` and `terminated` events.

    Attributes:
        completionThreshold (float): Consists of the percentage of media that should
            be consumed to trigger a completion.
        length (float): Consists of the length of the video.
    """

    length: NonNegativeFloat = Field(alias=CONTEXT_EXTENSION_LENGTH)
    completionThreshold: Optional[float] = Field(
        alias=CONTEXT_EXTENSION_COMPLETION_THRESHOLD
    )


class VideoEnableClosedCaptioningContextExtensions(VideoContextExtensions):
    """Represents the context.extensions field for video `interacted` xAPI statement.

    Attributes:
        ccSubtitleLanguage (str): Consists of the language of subtitle or closed
            captioning.
    """

    ccSubtitleLanguage: str = Field(alias=CONTEXT_EXTENSION_CC_SUBTITLE_LANG)


class VideoVolumeChangeInteractionContextExtensions(VideoContextExtensions):
    """Pydantic model for video volume change interaction `context`.`extensions`
    property.

    Attributes:
        volume (int): Consists of the volume of the video.
    """  # noqa: D205

    volume: int = Field(alias=CONTEXT_EXTENSION_VOLUME)


class VideoScreenChangeInteractionContextExtensions(VideoContextExtensions):
    """Pydantic model for video screen change interaction `context`.`extensions`
    property.

    Attributes:
        fullScreen (bool): Indicates whether the video is played in full screen mode.
        screenSize (str): Expresses the total available screen size for Video playback.
        videoPlaybackSize (str): Consists of the size in Width x Height of the video as
            viewed by the user.
    """  # noqa: D205

    fullScreen: bool = Field(alias=CONTEXT_EXTENSION_FULL_SCREEN)
    screenSize: str = Field(alias=CONTEXT_EXTENSION_SCREEN_SIZE)
    videoPlaybackSize: str = Field(alias=CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE)


class VideoInitializedContext(BaseVideoContext):
    """Pydantic model for video initialized `context` property.

    Attributes:
        extensions (dict): See VideoInitializedContextExtensions.
    """

    extensions: VideoInitializedContextExtensions


class VideoPlayedContext(BaseVideoContext):
    """Pydantic model for video played `context` property.

    Attributes:
        extensions (dict): See VideoContextExtensions.
    """

    extensions: Optional[VideoContextExtensions]


class VideoPausedContext(BaseVideoContext):
    """Pydantic model for video paused `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoSeekedContext(BaseVideoContext):
    """Pydantic model for video seeked `context` property.

    Attributes:
        extensions (dict): See VideoContextExtensions.
    """

    extensions: Optional[VideoContextExtensions]


class VideoCompletedContext(BaseVideoContext):
    """Pydantic model for video completed `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoTerminatedContext(BaseVideoContext):
    """Pydantic model for video terminated `context` property.

    Attributes:
        extensions (dict): See VideoBrowsingContextExtensions.
    """

    extensions: VideoBrowsingContextExtensions


class VideoEnableClosedCaptioningContext(BaseVideoContext):
    """Pydantic model for video enable closed captioning `context` property.

    Attributes:
        extensions (dict): See VideoEnableClosedCaptioningContextExtensions.
    """

    extensions: VideoEnableClosedCaptioningContextExtensions


class VideoVolumeChangeInteractionContext(BaseVideoContext):
    """Pydantic model for video volume change interaction `context` property.

    Attributes:
        extensions (dict): See VideoVolumeChangeInteractionContextExtensions.
    """

    extensions: VideoVolumeChangeInteractionContextExtensions


class VideoScreenChangeInteractionContext(BaseVideoContext):
    """Pydantic model for video screen change interaction `context` property.

    Attributes:
        extensions (dict): See VideoScreenChangeInteractionContextExtensions.
    """

    extensions: VideoScreenChangeInteractionContextExtensions



"""API routes related to application health checking."""

import logging
from typing import Union

from fastapi import APIRouter, Response, status
from pydantic import BaseModel

from ralph.backends.data.base import DataBackendStatus
from ralph.backends.loader import get_lrs_backends
from ralph.backends.lrs.base import BaseAsyncLRSBackend, BaseLRSBackend
from ralph.conf import settings
from ralph.utils import await_if_coroutine, get_backend_class

logger = logging.getLogger(__name__)

router = APIRouter()

BACKEND_CLIENT: Union[BaseLRSBackend, BaseAsyncLRSBackend] = get_backend_class(
    backends=get_lrs_backends(), name=settings.RUNSERVER_BACKEND
)()


class Heartbeat(BaseModel):
    """Ralph backends status."""

    database: DataBackendStatus

    @property
    def is_alive(self):
        """A helper that checks the overall status."""
        return self.database == DataBackendStatus.OK


@router.get("/__lbheartbeat__")
async def lbheartbeat() -> None:
    """Load balancer heartbeat.

    Return a 200 when the server is running.
    """
    return


@router.get("/__heartbeat__", status_code=status.HTTP_200_OK)
async def heartbeat(response: Response) -> Heartbeat:
    """Application heartbeat.

    Return a 200 if all checks are successful.
    """
    statuses = Heartbeat.construct(
        database=await await_if_coroutine(BACKEND_CLIENT.status())
    )
    if not statuses.is_alive:
        response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    return statuses



"""Tests for ralph.parsers module."""

import gzip
import logging
import shutil
from io import StringIO

import pytest

from ralph.conf import settings
from ralph.parsers import GELFParser


def test_parsers_gelfparser_parse_empty_file():
    """Test the GELFParser parsing with an empty file."""
    parser = GELFParser()

    empty_file = StringIO()
    with pytest.raises(StopIteration):
        next(parser.parse(empty_file))


def test_parsers_gelfparser_parse_raw_file(gelf_logger):
    """Test the GELFParser parsing using a flat GELF file."""
    parser = GELFParser()

    # Create two log entries
    gelf_logger.info('{"username": "foo"}')
    gelf_logger.info('{"username": ""}')

    with open(
        gelf_logger.handlers[0].stream.name, encoding=settings.LOCALE_ENCODING
    ) as stream:
        events = list(parser.parse(stream))

    assert len(events) == 2
    assert events[0] == '{"username": "foo"}'
    assert events[1] == '{"username": ""}'


def test_parsers_gelfparser_parse_gzipped_file(fs, gelf_logger):
    """Test the GELFParser parsing using a gzipped GELF file."""
    gelf_logger.info('{"username": "foo"}')
    gelf_logger.info('{"username": "bar"}')

    # Compress the log file
    log_file_name = gelf_logger.handlers[0].stream.name
    gzipped_log_file_name = f"{log_file_name}.gz"
    with open(log_file_name, "rb") as log_file:
        with gzip.open(gzipped_log_file_name, "wb") as gzipped_log_file:
            shutil.copyfileobj(log_file, gzipped_log_file)

    parser = GELFParser()
    with open(
        gelf_logger.handlers[0].stream.name, encoding=settings.LOCALE_ENCODING
    ) as stream:
        events = list(parser.parse(stream))

    assert len(events) == 2
    assert events[0] == '{"username": "foo"}'
    assert events[1] == '{"username": "bar"}'


def test_parsers_gelfparser_parse_partially_invalid_file(caplog):
    """Test the GELFParser with a file containing invalid JSON strings."""
    with StringIO() as file:
        file.writelines(
            [
                # This is invalid gelf, but we assume it's valid in our case
                '{"short_message": "This seems valid."}\n',
                # Invalid json
                "{ This is not valid json and raises json.decoder.JSONDecodeError\n",
                # Valid json but invalid gelf raises KeyError:
                # key "short_message" not found
                "{}\n",
                # As above but raises TypeError:
                # list indices must be integers or slices, not str
                "[]\n",
                # Another assumed valid gelf
                '{"short_message": {"username": "This seems valid too."}}\n',
            ]
        )

        file.seek(0)
        parser = GELFParser()
        events = list(parser.parse(file))

    assert len(events) == 2
    assert events[0] == "This seems valid."
    assert events[1] == {"username": "This seems valid too."}

    caplog.clear()
    with StringIO() as file:
        file.write("{ This is not valid json and raises json.decoder.JSONDecodeError\n")
        file.seek(0)
        parser = GELFParser()
        with caplog.at_level(logging.DEBUG):
            events = list(parser.parse(file))

    assert len(events) == 0
    assert (
        "ralph.parsers",
        logging.ERROR,
        "Input event '{ This is not valid json and raises "
        "json.decoder.JSONDecodeError\n' "
        "is not a valid JSON string! It will be ignored.",
    ) in caplog.record_tuples
    assert (
        "ralph.parsers",
        logging.DEBUG,
        "Raised error was: Expecting property name enclosed in double quotes: "
        "line 1 column 3 (char 2)",
    ) in caplog.record_tuples

    caplog.clear()
    with StringIO() as file:
        file.write("{}")
        file.seek(0)
        parser = GELFParser()
        with caplog.at_level(logging.DEBUG):
            events = list(parser.parse(file))

    assert len(events) == 0
    assert (
        "ralph.parsers",
        logging.ERROR,
        "Input event '{}' doesn't comply with GELF format! It will be ignored.",
    ) in caplog.record_tuples
    assert (
        "ralph.parsers",
        logging.DEBUG,
        "Raised error was: 'short_message'",
    ) in caplog.record_tuples

    caplog.clear()
    with StringIO() as file:
        file.write("[]")
        file.seek(0)
        parser = GELFParser()
        with caplog.at_level(logging.DEBUG):
            events = list(parser.parse(file))

    assert len(events) == 0
    assert (
        "ralph.parsers",
        logging.ERROR,
        "Input event '[]' is not a valid JSON string! It will be ignored.",
    ) in caplog.record_tuples
    assert (
        "ralph.parsers",
        logging.DEBUG,
        "Raised error was: list indices must be integers or slices, not str",
    ) in caplog.record_tuples



"""Ralph exceptions."""


class BackendException(Exception):
    """Raised when a backend has a failure."""


class BackendParameterException(Exception):
    """Raised when a backend parameter value is not valid."""


class BadFormatException(Exception):
    """Raised when the format of an event is not valid."""


class ConfigurationException(Exception):
    """Raised when the configuration is not valid."""


class ConversionException(Exception):
    """Raised when a converter has a failure."""


class EventKeyError(Exception):
    """Raised when an expected event key has not been found."""


class MissingConversionSetException(Exception):
    """Raised when an expected conversion set has not been found."""


class UnknownEventException(Exception):
    """Raised when no pydantic model is found for a given event."""


class UnsupportedBackendException(Exception):
    """Raised when trying to use an unsupported backend type."""



"""MongoDB data backend for Ralph."""

from __future__ import annotations

import hashlib
import logging
import struct
from io import IOBase
from typing import Generator, Iterable, Iterator, List, Optional, Tuple, TypeVar, Union
from uuid import uuid4

from bson.errors import BSONError
from bson.objectid import ObjectId
from dateutil.parser import isoparse
from pydantic import MongoDsn, PositiveInt, constr
from pymongo import MongoClient, ReplaceOne
from pymongo.collection import Collection
from pymongo.errors import (
    BulkWriteError,
    ConnectionFailure,
    InvalidName,
    InvalidOperation,
    PyMongoError,
)

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    BaseQuery,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.conf import BaseSettingsConfig, ClientOptions
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import iter_by_batch

logger = logging.getLogger(__name__)


class MongoClientOptions(ClientOptions):
    """MongoDB additional client options."""

    document_class: Optional[str] = None
    tz_aware: Optional[bool] = None


class MongoDataBackendSettings(BaseDataBackendSettings):
    """MongoDB data backend default configuration.

    Attributes:
        CONNECTION_URI (str): The MongoDB connection URI.
        DEFAULT_DATABASE (str): The MongoDB database to connect to.
        DEFAULT_COLLECTION (str): The MongoDB database collection to get objects from.
        CLIENT_OPTIONS (MongoClientOptions): A dictionary of MongoDB client options.
        LOCALE_ENCODING (str): The locale encoding to use when none is provided.
        READ_CHUNK_SIZE (int): The default chunk size for reading batches of documents.
        WRITE_CHUNK_SIZE (int): The default chunk size for writing batches of documents.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__MONGO__"

    CONNECTION_URI: MongoDsn = MongoDsn("mongodb://localhost:27017/", scheme="mongodb")
    DEFAULT_DATABASE: constr(regex=r"^[^\s.$/\\\"\x00]+$") = "statements"
    DEFAULT_COLLECTION: constr(
        regex=r"^(?!.*\.\.)[^.$\x00]+(?:\.[^.$\x00]+)*$"
    ) = "marsha"
    CLIENT_OPTIONS: MongoClientOptions = MongoClientOptions()


class MongoQuery(BaseQuery):
    """MongoDB query model.

    Attributes:
        filter (dict): A filter query to select which documents to include.
        limit (int): The maximum number of results to return.
        projection (dict): Dictionary specifying the fields to include or exclude.
        sort (list): A list of (key, direction) pairs specifying the sort order.
    """

    filter: Optional[dict] = None
    limit: Optional[int] = None
    projection: Optional[dict] = None
    sort: Optional[List[Tuple]] = None


Settings = TypeVar("Settings", bound=MongoDataBackendSettings)


class MongoDataBackend(BaseDataBackend[Settings, MongoQuery], Writable, Listable):
    """MongoDB data backend."""

    name = "mongo"
    unsupported_operation_types = {BaseOperationType.APPEND}

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the MongoDB client.

        Args:
            settings (MongoDataBackendSettings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self.client = MongoClient(
            self.settings.CONNECTION_URI, **self.settings.CLIENT_OPTIONS.dict()
        )
        self.database = self.client[self.settings.DEFAULT_DATABASE]
        self.collection = self.database[self.settings.DEFAULT_COLLECTION]

    def status(self) -> DataBackendStatus:
        """Check the MongoDB connection status.

        Return:
            DataBackendStatus: The status of the data backend.
        """
        # Check MongoDB connection.
        try:
            self.client.admin.command("ping")
        except (ConnectionFailure, InvalidOperation) as error:
            logger.error("Failed to connect to MongoDB: %s", error)
            return DataBackendStatus.AWAY

        # Check MongoDB server status.
        try:
            if (
                self.client.admin.command("serverStatus").get("ok")
                != 1.0  # noqa: PLR2004
            ):
                logger.error("MongoDB `serverStatus` command did not return 1.0")
                return DataBackendStatus.ERROR
        except PyMongoError as error:
            logger.error("Failed to get MongoDB server status: %s", error)
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List collections in the `target` database.

        Args:
            target (str or None): The MongoDB database name to list collections from.
                If target is `None`, the `DEFAULT_DATABASE` is used instead.
            details (bool): Get detailed collection information instead of just IDs.
            new (bool): Ignored.

        Raise:
            BackendException: If a failure during the list operation occurs.
            BackendParameterException: If the `target` is not a valid database name.
        """
        if new:
            logger.warning("The `new` argument is ignored")

        try:
            database = self.client[target] if target else self.database
        except InvalidName as error:
            msg = "The target=`%s` is not a valid database name: %s"
            logger.error(msg, target, error)
            raise BackendParameterException(msg % (target, error)) from error

        try:
            for collection_info in database.list_collections():
                if details:
                    yield collection_info
                else:
                    yield collection_info.get("name")
        except PyMongoError as error:
            msg = "Failed to list MongoDB collections: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def read(  # noqa: PLR0913
        self,
        query: Optional[MongoQuery] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read documents matching the `query` from `target` collection and yield them.

        Args:
            query (MongoQuery): The MongoDB query to use when reading documents.
            target (str or None): The MongoDB collection name to query.
                If target is `None`, the `DEFAULT_COLLECTION` is used instead.
            chunk_size (int or None): The chunk size when reading archives by batch.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Whether to yield dictionaries or bytes.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a failure occurs during MongoDB connection or
                during encoding documents and `ignore_errors` is set to `False`.
            BackendParameterException: If the `target` is not a valid collection name.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_dicts(
        self,
        query: MongoQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        kwargs = query.dict(exclude_unset=True)
        collection = self._get_target_collection(target)
        try:
            documents = collection.find(batch_size=chunk_size, **kwargs)
            yield from (d.update({"_id": str(d.get("_id"))}) or d for d in documents)
        except (PyMongoError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to execute MongoDB query: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write `data` documents to the `target` collection and return their count.

        Args:
            data (Iterable or IOBase): The data containing documents to write.
            target (str or None): The target MongoDB collection name.
            chunk_size (int or None): The number of documents to write in one batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool):  If `True`, errors during decoding, encoding and
                sending batches of documents are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                    instead. See `BaseOperationType`.

        Return:
            int: The number of documents written.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If the `operation_type` is `APPEND` as it is not
                supported.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        count = 0
        collection = self._get_target_collection(target)
        msg = "Start writing to the %s collection of the %s database (chunk size: %d)"
        logger.debug(msg, collection, self.database, chunk_size)
        if operation_type == BaseOperationType.UPDATE:
            for batch in iter_by_batch(self.to_replace_one(data), chunk_size):
                count += self._bulk_update(batch, ignore_errors, collection)
            logger.info("Updated %d documents with success", count)
        elif operation_type == BaseOperationType.DELETE:
            for batch in iter_by_batch(self.to_ids(data), chunk_size):
                count += self._bulk_delete(batch, ignore_errors, collection)
            logger.info("Deleted %d documents with success", count)
        else:
            data = self.to_documents(data, ignore_errors, operation_type)
            for batch in iter_by_batch(data, chunk_size):
                count += self._bulk_import(batch, ignore_errors, collection)
            logger.info("Inserted %d documents with success", count)

        return count

    def close(self) -> None:
        """Close the MongoDB backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        try:
            self.client.close()
        except PyMongoError as error:
            msg = "Failed to close MongoDB client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    @staticmethod
    def to_ids(data: Iterable[dict]) -> Iterable[str]:
        """Convert `data` statements to ids."""
        for statement in data:
            yield statement.get("id")

    @staticmethod
    def to_replace_one(data: Iterable[dict]) -> Iterable[ReplaceOne]:
        """Convert `data` statements to Mongo `ReplaceOne` objects."""
        for statement in data:
            yield ReplaceOne(
                {"_source.id": {"$eq": statement.get("id")}},
                {"_source": statement},
            )

    @staticmethod
    def to_documents(
        data: Iterable[dict],
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> Generator[dict, None, None]:
        """Convert `data` statements to MongoDB documents.

        We expect statements to have at least an `id` and a `timestamp` field that will
        be used to compute a unique MongoDB Object ID. This ensures that we will not
        duplicate statements in our database and allows us to support pagination.
        """
        for statement in data:
            if "id" not in statement and operation_type == BaseOperationType.INDEX:
                msg = "statement %s has no 'id' field"
                if ignore_errors:
                    logger.warning("statement %s has no 'id' field", statement)
                    continue
                logger.error(msg, statement)
                raise BackendException(msg % statement)
            if "timestamp" not in statement:
                msg = "statement %s has no 'timestamp' field"
                if ignore_errors:
                    logger.warning(msg, statement)
                    continue
                logger.error(msg, statement)
                raise BackendException(msg % statement)
            try:
                timestamp = int(isoparse(statement["timestamp"]).timestamp())
            except ValueError as err:
                msg = "statement %s has an invalid 'timestamp' field"
                if ignore_errors:
                    logger.warning(msg, statement)
                    continue
                logger.error(msg, statement)
                raise BackendException(msg % statement) from err
            document = {
                "_id": ObjectId(
                    # This might become a problem in February 2106.
                    # Meanwhile, we use the timestamp in the _id field for pagination.
                    struct.pack(">I", timestamp)
                    + bytes.fromhex(
                        hashlib.sha256(
                            bytes(statement.get("id", str(uuid4())), "utf-8")
                        ).hexdigest()[:16]
                    )
                ),
                "_source": statement,
            }

            yield document

    def _get_target_collection(self, target: Union[str, None]) -> Collection:
        """Return the validated target collection."""
        try:
            return self.database[target] if target else self.collection
        except InvalidName as error:
            msg = "The target=`%s` is not a valid collection name: %s"
            logger.error(msg, target, error)
            raise BackendParameterException(msg % (target, error)) from error

    def _bulk_import(
        self, batch: List, ignore_errors: bool, collection: Collection
    ) -> int:
        """Insert a `batch` of documents into the MongoDB `collection`."""
        try:
            new_documents = collection.insert_many(batch)
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to insert document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nInserted", 0)
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        inserted_count = len(new_documents.inserted_ids)
        logger.debug("Inserted %d documents chunk with success", inserted_count)
        return inserted_count

    def _bulk_delete(
        self, batch: List, ignore_errors: bool, collection: Collection
    ) -> int:
        """Delete a `batch` of documents from the MongoDB `collection`."""
        try:
            deleted_documents = collection.delete_many({"_source.id": {"$in": batch}})
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to delete document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nRemoved", 0)
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        deleted_count = deleted_documents.deleted_count
        logger.debug("Deleted %d documents chunk with success", deleted_count)
        return deleted_count

    def _bulk_update(
        self, batch: List, ignore_errors: bool, collection: Collection
    ) -> int:
        """Update a `batch` of documents into the MongoDB `collection`."""
        try:
            updated_documents = collection.bulk_write(batch)
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to update document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nModified", 0)
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        modified_count = updated_documents.modified_count
        logger.debug("Updated %d documents chunk with success", modified_count)
        return modified_count



"""Base xAPI `Context` definitions."""

from typing import Dict, List, Optional, Union
from uuid import UUID

from pydantic import StrictStr

from ..config import BaseModelWithConfig
from .agents import BaseXapiAgent
from .common import IRI, LanguageTag
from .groups import BaseXapiGroup
from .unnested_objects import BaseXapiActivity, BaseXapiStatementRef


class BaseXapiContextContextActivities(BaseModelWithConfig):
    """Pydantic model for context `contextActivities` property.

    Attributes:
        parent (dict or list): An Activity with a direct relation to the statement's
            Activity.
        grouping (dict or list): An Activity with an indirect relation to the
            statement's Activity.
        category (dict or list): An Activity used to categorize the Statement.
        other (dict or list): A contextActivity that doesn't fit one of the other
            properties.
    """

    parent: Optional[Union[BaseXapiActivity, List[BaseXapiActivity]]]
    grouping: Optional[Union[BaseXapiActivity, List[BaseXapiActivity]]]
    category: Optional[Union[BaseXapiActivity, List[BaseXapiActivity]]]
    other: Optional[Union[BaseXapiActivity, List[BaseXapiActivity]]]


class BaseXapiContext(BaseModelWithConfig):
    """Pydantic model for `context` property.

    Attributes:
        registration (UUID): The registration that the Statement is associated with.
        instructor (dict): The instructor that the Statement relates to.
        team (dict): The team that this Statement relates to.
        contextActivities (dict): See BaseXapiContextContextActivities.
        revision (str): The revision of the activity associated with this Statement.
        platform (str): The platform where the learning activity took place.
        language (dict): The language in which the experience occurred.
        statement (dict): Another Statement giving context for this Statement.
        extensions (dict): Consists of a dictionary of other properties as needed.
    """

    registration: Optional[UUID]
    instructor: Optional[BaseXapiAgent]
    team: Optional[BaseXapiGroup]
    contextActivities: Optional[BaseXapiContextContextActivities]
    revision: Optional[StrictStr]
    platform: Optional[StrictStr]
    language: Optional[LanguageTag]
    statement: Optional[BaseXapiStatementRef]
    extensions: Optional[Dict[IRI, Union[str, int, bool, list, dict, None]]]



"""Textbook interaction event fields definitions."""

import sys
from typing import Optional, Union

from pydantic import Field, constr

from ...base import AbstractBaseEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class TextbookInteractionBaseEventField(AbstractBaseEventField):
    """Pydantic model for textbook interaction core `event` field.

    Attributes:
        chapter (str): Consists of the name of the PDF file.
            It begins with the `block_id` value and ends with the `.pdf` extension.
        page (int): The number of the page that is open when the event is emitted.
    """

    page: int
    chapter: constr(
        regex=(r"^\/asset-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+type@asset\+block.+$")
    )


class TextbookPdfThumbnailsToggledEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.thumbnails.toggled`.`event` field.

    Attribute:
        name (str): Consists of the value `textbook.pdf.thumbnails.toggled`.
    """

    name: Literal["textbook.pdf.thumbnails.toggled"]


class TextbookPdfThumbnailNavigatedEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.thumbnail.navigated`.`event` field.

    Attribute:
        name (str): Consists of the value `textbook.pdf.thumbnail.navigated`.
        thumbnail_title (str): Consists of the name of the thumbnail.
    """

    name: Literal["textbook.pdf.thumbnail.navigated"]
    thumbnail_title: str


class TextbookPdfOutlineToggledEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.outline.toggled`.`event` field.

    Attribute:
        name (str): Consists of the value `textbook.pdf.outline.toggled`.
    """

    name: Literal["textbook.pdf.outline.toggled"]


class TextbookPdfChapterNavigatedEventField(AbstractBaseEventField):
    """Pydantic model for `textbook.pdf.chapter.navigated`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.chapter.navigated`.
        chapter (str): Consists of the name of the PDF file.
            It begins with the `block_id` value and ends with the `.pdf` extension.
    """

    name: Literal["textbook.pdf.chapter.navigated"]
    chapter: constr(
        regex=(r"^\/asset-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+type@asset\+block.+$")
    )
    chapter_title: str


class TextbookPdfPageNavigatedEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.page.navigated`.`event` field.

    Attribute:
        name (str): Consists of the value `textbook.pdf.page.navigated`.
    """

    name: Literal["textbook.pdf.page.navigated"]


class TextbookPdfZoomButtonsChangedEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.zoom.buttons.changed`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.zoom.buttons.changed`.
        direction (str): Consists of either the `in` or `out` value.
    """

    name: Literal["textbook.pdf.zoom.buttons.changed"]
    direction: Union[Literal["in"], Literal["out"]]


class TextbookPdfZoomMenuChangedEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.zoom.menu.changed`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.zoom.menu.changed`.
        amount (str): Consists either of the `0.5`, `0.75`, `1`, `1.25`, `1.5`, `2`,
            `3`, `4`, `auto`, `custom`, `page-actual`, `page-fit`, `page-width` value.
    """

    name: Literal["textbook.pdf.zoom.menu.changed"]
    amount: Union[
        Literal["0.5"],
        Literal["0.75"],
        Literal["1"],
        Literal["1.25"],
        Literal["1.5"],
        Literal["2"],
        Literal["3"],
        Literal["4"],
        Literal["auto"],
        Literal["custom"],
        Literal["page-actual"],
        Literal["page-fit"],
        Literal["page-width"],
    ]


class TextbookPdfDisplayScaledEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.display.scaled`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.display.scaled`.
        amount (str): Consists of a floating point number string value.
    """

    name: Literal["textbook.pdf.display.scaled"]
    amount: float


class TextbookPdfPageScrolledEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.page.scrolled`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.page.scrolled`.
        direction (str): Consists either of the `up` or `down` value.
    """

    name: Literal["textbook.pdf.page.scrolled"]
    direction: Union[Literal["up"], Literal["down"]]


class TextbookPdfSearchExecutedEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.search.executed`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.search.executed`.
        caseSensitive (bool): Consists either of the `true` value if the case-sensitive
            option is selected or `false` if this option is not selected.
        highlightAll (bool): Consists either of the `true` value if the option to
            highlight all matches is selected or `false` if this option is not selected.
        query (str): Consists of the value in the search field.
        status (str): Consists either of the value `not found` for a search string that
            is unsuccessful or blank for successful search strings.
    """

    name: Literal["textbook.pdf.search.executed"]
    caseSensitive: bool
    highlightAll: bool
    query: str
    status: str


class TextbookPdfSearchNavigatedNextEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.search.navigatednext`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.search.navigatednext`.
        caseSensitive (bool): Consists either of the `true` value if the case-sensitive
            option is selected or `false` if this option is not selected.
        findPrevious(bool): Consists either of the ‘true’ value if the user clicks the
            Find Previous icon or ‘false’ if the user clicks the <kbd>Find Next</kbd>
            icon.
        highlightAll (bool): Consists either of the `true` value if the option to
            highlight all matches is selected or `false` if this option is not selected.
        query (str): Consists of the value in the search field.
        status (str): Consists either of the value `not found` for a search string that
            is unsuccessful or blank for successful search strings.
    """

    name: Literal["textbook.pdf.search.navigatednext"]
    caseSensitive: bool
    findPrevious: bool
    highlightAll: bool
    query: str
    status: str


class TextbookPdfSearchHighlightToggledEventField(TextbookInteractionBaseEventField):
    """Pydantic model for `textbook.pdf.search.highlight.toggled`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.search.highlight.toggled`.
        caseSensitive (bool): Consists either of the `true` value if the case sensitive
            option is selected or `false` if this option is not selected.
        highlightAll (bool): Consists either of the `true` value if the option to
            highlight all matches is selected or `false` if this option is not selected.
        query (str): Consists of the value in the search field.
        status (str): Consists either of the value `not found` for a search string that
            is unsuccessful or blank for successful search strings.
    """

    name: Literal["textbook.pdf.search.highlight.toggled"]
    caseSensitive: bool
    highlightAll: bool
    query: str
    status: str


class TextbookPdfSearchCaseSensitivityToggledEventField(
    TextbookInteractionBaseEventField
):
    """Pydantic model for `textbook.pdf.searchcasesensitivity.toggled`.`event` field.

    Attributes:
        name (str): Consists of the value `textbook.pdf.searchcasesensitivity.toggled`.
        caseSensitive (bool): Consists either of the `true` value if the case sensitive
            option is selected or `false` if this option is not selected.
        highlightAll (bool): Consists either of the `true` value if the option to
            highlight all matches is selected or `false` if this option is not selected.
        query (str): Consists of the value in the search field.
        status (str): Consists either of the value `not found` for a search string that
            is unsuccessful or blank for successful search strings.
    """

    name: Literal["textbook.pdf.searchcasesensitivity.toggled"]
    caseSensitive: bool
    highlightAll: bool
    query: str
    status: str


class BookEventField(AbstractBaseEventField):
    """Pydantic model for `book`.`event` field.

    Attributes:
        chapter (str): Consists of the name of the PDF file.
        name (str): Consists of `textbook.pdf.page.loaded` if type is set to
            `gotopage`,
            `textbook.pdf.page.navigatednext` if type is set to `prevpage`,
            `textbook.pdf.page.navigatednext` if type is set to `nextpage`.
        new (int): Consists of the destination page number.
        old (int): Consists of the original page number. It applies to `gotopage` event
            types only.
        type (str): Consists of `gotopage` value when a page loads after the student
            manually enters its number, `prevpage` value when the next page button is
            clicked or `nextpage` value when the previous page button is clicked.
    """

    chapter: constr(
        regex=(r"^\/asset-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+type@asset\+block.+$")
    )
    name: Union[
        Literal["textbook.pdf.page.loaded"], Literal["textbook.pdf.page.navigatednext"]
    ]
    new: int
    old: Optional[int]
    type: Union[Literal["gotopage"], Literal["prevpage"], Literal["nextpage"]] = Field(
        alias="type"
    )



"""Tests for the ralph.logger module."""

import pytest
from click.testing import CliRunner

import ralph.logger
from ralph.cli import cli
from ralph.conf import settings
from ralph.exceptions import ConfigurationException


def test_logger_exists(fs, monkeypatch):
    """Test the logging system when a correct configuration is provided."""
    mock_default_config = {
        "version": 1,
        "propagate": True,
        "formatters": {
            "ralph_format": {
                "format": "%(asctime)-23s %(levelname)-8s %(name)-8s %(message)s"
            },
        },
        "handlers": {
            "ralph_stderr": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "stream": "ext://sys.stderr",
                "formatter": "ralph_format",
            },
        },
        "loggers": {
            "ralph": {
                "handlers": ["ralph_stderr"],
                "level": "DEBUG",
            },
        },
    }

    fs.create_dir(str(settings.APP_DIR))
    fs.create_dir("foo")

    monkeypatch.setattr(ralph.logger.settings, "LOGGING", mock_default_config)

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["write", "-b", "fs", "-t", "test_file", "--fs-default-directory-path", "foo"],
        input="test input",
    )

    assert result.exit_code == 0
    assert "Writing to target test_file for the configured fs backend" in result.output
    assert "Backend parameters:" in result.output


def test_logger_no_config(fs, monkeypatch):
    """Test that an error occurs when no logging configuration exists."""
    mock_default_config = None

    monkeypatch.setattr(ralph.logger.settings, "LOGGING", mock_default_config)

    runner = CliRunner()

    with pytest.raises(ConfigurationException):
        result = runner.invoke(cli, ["list", "-b", "fs"], catch_exceptions=False)
        assert result.exit_code == 1


def test_logger_bad_config(fs, monkeypatch):
    """Test that an error occurs when a logging is improperly configured."""
    mock_default_config = "this is not a valid json"

    monkeypatch.setattr(ralph.logger.settings, "LOGGING", mock_default_config)

    runner = CliRunner()

    with pytest.raises(ConfigurationException):
        result = runner.invoke(cli, ["list", "-b", "fs"], catch_exceptions=False)
        assert result.exit_code == 1



"""Navigational event model definitions."""

import sys
from typing import Union

from pydantic import Json, validator

from ralph.models.selector import selector

from ..browser import BaseBrowserModel
from .fields.events import NavigationalEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class UIPageClose(BaseBrowserModel):
    """Pydantic model for `page_close` statement.

    The browser emits this statement when the user navigates to the next page
    or closes the browser window (when the JavaScript `window.onunload` event
    is called).

    Attributes:
        event (str): Consists of the string value `{}`.
        event_type (str): Consists of the value `page_close`.
        name (str): Consists of the value `page_close`.
    """

    __selector__ = selector(event_source="browser", event_type="page_close")

    event: Literal["{}"]
    event_type: Literal["page_close"]
    name: Literal["page_close"]


class UISeqGoto(BaseBrowserModel):
    """Pydantic model for `seq_goto` statement.

    The browser emits this statement when a user jumps between units in a sequence.

    Attributes:
        event (obj): Consists of member fields that identify specifics triggered event.
        event_type (str): Consists of the value `seq_goto`.
        name (str): Consists of the value `seq_goto`.
    """

    __selector__ = selector(event_source="browser", event_type="seq_goto")

    event: Union[Json[NavigationalEventField], NavigationalEventField]
    event_type: Literal["seq_goto"]
    name: Literal["seq_goto"]


class UISeqNext(BaseBrowserModel):
    """Pydantic model for `seq_next` statement.

    The browser emits this statement when a user navigates to the next unit in a
    sequence.

    Attributes:
        event (obj): Consists of member fields that identify specifics triggered event.
        event_type (str): Consists of the value `seq_next`.
        name (str): Consists of the value `seq_next`.
    """

    __selector__ = selector(event_source="browser", event_type="seq_next")

    event: Union[Json[NavigationalEventField], NavigationalEventField]
    event_type: Literal["seq_next"]
    name: Literal["seq_next"]

    @validator("event")
    @classmethod
    def validate_next_jump_event_field(
        cls, value: Union[Json[NavigationalEventField], NavigationalEventField]
    ) -> Union[Json[NavigationalEventField], NavigationalEventField]:
        """Check that event.new is equal to event.old + 1."""
        if value.new != value.old + 1:
            raise ValueError("event.new - event.old should be equal to 1")

        return value


class UISeqPrev(BaseBrowserModel):
    """Pydantic model for `seq_prev` statement.

    The browser emits this statement when a user navigates to the previous unit in a
    sequence.

    Attributes:
        event (obj): Consists of member fields that identify specifics triggered event.
        event_type (str): Consists of the value `seq_prev`.
        name (str): Consists of the value `seq_prev`.

    """

    __selector__ = selector(event_source="browser", event_type="seq_prev")

    event: Union[Json[NavigationalEventField], NavigationalEventField]
    event_type: Literal["seq_prev"]
    name: Literal["seq_prev"]

    @validator("event")
    @classmethod
    def validate_prev_jump_event_field(
        cls, value: Union[Json[NavigationalEventField], NavigationalEventField]
    ) -> Union[Json[NavigationalEventField], NavigationalEventField]:
        """Check that event.new is equal to event.old - 1."""
        if value.new != value.old - 1:
            raise ValueError("event.old - event.new should be equal to 1")

        return value



"""Enrollment event xAPI Converter."""
from typing import Set

from ralph.models.converter import ConversionItem
from ralph.models.edx.enrollment.statements import (
    EdxCourseEnrollmentActivated,
    EdxCourseEnrollmentDeactivated,
)
from ralph.models.xapi.lms.statements import LMSRegisteredCourse, LMSUnregisteredCourse

from .base import BaseXapiConverter


class LMSBaseXapiConverter(BaseXapiConverter):
    """Base LMS xAPI Converter."""

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "object__id",
                    "event__course_id",
                    lambda course_id: f"{self.platform_url}/courses/{course_id}/info",
                ),
                ConversionItem(
                    "context__contextActivities__category",
                    None,
                    lambda _: [{"id": "https://w3id.org/xapi/lms"}],
                ),
            },
        )


class EdxCourseEnrollmentActivatedToLMSRegisteredCourse(LMSBaseXapiConverter):
    """Convert a common edX `edx.course.enrollment.activated` event to xAPI."""

    __src__ = EdxCourseEnrollmentActivated
    __dest__ = LMSRegisteredCourse


class EdxCourseEnrollmentDeactivatedToLMSUnregisteredCourse(LMSBaseXapiConverter):
    """Convert a common edX `edx.course.enrollment.deactivated` event to xAPI."""

    __src__ = EdxCourseEnrollmentDeactivated
    __dest__ = LMSUnregisteredCourse



"""Elasticsearch LRS backend for Ralph."""

import logging
from typing import Iterator, List, Optional

from ralph.backends.data.es import (
    ESDataBackend,
    ESDataBackendSettings,
    ESQuery,
    ESQueryPit,
)
from ralph.backends.lrs.base import (
    AgentParameters,
    BaseLRSBackend,
    BaseLRSBackendSettings,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException

logger = logging.getLogger(__name__)


class ESLRSBackendSettings(BaseLRSBackendSettings, ESDataBackendSettings):
    """Elasticsearch LRS backend default configuration."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__LRS__ES__"


class ESLRSBackend(BaseLRSBackend[ESLRSBackendSettings], ESDataBackend):
    """Elasticsearch LRS backend implementation."""

    def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""
        query = self.get_query(params=params)
        try:
            es_documents = self.read(
                query=query, target=target, chunk_size=params.limit
            )
            statements = [document["_source"] for document in es_documents]
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from Elasticsearch")
            raise error

        return StatementQueryResult(
            statements=statements,
            pit_id=query.pit.id,
            search_after="|".join(query.search_after) if query.search_after else "",
        )

    def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> Iterator[dict]:
        """Yield statements with matching ids from the backend."""
        query = self.query_class(query={"terms": {"_id": ids}})
        try:
            es_response = self.read(query=query, target=target)
            yield from (document["_source"] for document in es_response)
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from Elasticsearch")
            raise error

    @staticmethod
    def get_query(params: RalphStatementsQuery) -> ESQuery:
        """Construct query from statement parameters."""
        es_query_filters = []

        if params.statement_id:
            es_query_filters += [{"term": {"_id": params.statement_id}}]

        ESLRSBackend._add_agent_filters(es_query_filters, params.agent, "actor")
        ESLRSBackend._add_agent_filters(es_query_filters, params.authority, "authority")

        if params.verb:
            es_query_filters += [{"term": {"verb.id.keyword": params.verb}}]

        if params.activity:
            es_query_filters += [
                {"term": {"object.objectType.keyword": "Activity"}},
                {"term": {"object.id.keyword": params.activity}},
            ]

        if params.since:
            es_query_filters += [{"range": {"timestamp": {"gt": params.since}}}]

        if params.until:
            es_query_filters += [{"range": {"timestamp": {"lte": params.until}}}]

        es_query = {
            "pit": ESQueryPit.construct(id=params.pit_id),
            "size": params.limit,
            "sort": [{"timestamp": {"order": "asc" if params.ascending else "desc"}}],
        }
        if len(es_query_filters) > 0:
            es_query["query"] = {"bool": {"filter": es_query_filters}}

        if params.ignore_order:
            es_query["sort"] = "_shard_doc"

        if params.search_after:
            es_query["search_after"] = params.search_after.split("|")

        # Note: `params` fields are validated thus we skip their validation in ESQuery.
        return ESQuery.construct(**es_query)

    @staticmethod
    def _add_agent_filters(
        es_query_filters: list, agent_params: AgentParameters, target_field: str
    ) -> None:
        """Add filters relative to agents to `es_query_filters`."""
        if not agent_params:
            return

        if not isinstance(agent_params, dict):
            agent_params = agent_params.dict()

        if agent_params.get("mbox"):
            field = f"{target_field}.mbox.keyword"
            es_query_filters += [{"term": {field: agent_params.get("mbox")}}]
        elif agent_params.get("mbox_sha1sum"):
            field = f"{target_field}.mbox_sha1sum.keyword"
            es_query_filters += [{"term": {field: agent_params.get("mbox_sha1sum")}}]
        elif agent_params.get("openid"):
            field = f"{target_field}.openid.keyword"
            es_query_filters += [{"term": {field: agent_params.get("openid")}}]
        elif agent_params.get("account__name"):
            field = f"{target_field}.account.name.keyword"
            es_query_filters += [{"term": {field: agent_params.get("account__name")}}]
            field = f"{target_field}.account.homePage.keyword"
            es_query_filters += [
                {"term": {field: agent_params.get("account__home_page")}}
            ]



"""Converter methods definition."""

import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib import import_module
from inspect import getmembers, isclass
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    Optional,
    Set,
    TextIO,
    Tuple,
    Union,
)

from pydantic import BaseModel, ValidationError

from ralph.conf import MODEL_PATH_SEPARATOR
from ralph.exceptions import (
    BadFormatException,
    ConversionException,
    MissingConversionSetException,
    UnknownEventException,
)
from ralph.utils import get_dict_value_from_path, set_dict_value_from_path

from .selector import ModelSelector

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ConversionItem:
    """Conversion set item."""

    dest: Tuple[str]
    src: Union[Tuple[str], str, None]
    transformers: Tuple[Callable[[Any], Any]]
    raw_input: bool

    def __init__(
        self,
        dest: str,
        src: Optional[str] = None,
        transformers=lambda _: _,
        raw_input: bool = False,
    ) -> None:
        """Initialize ConversionItem.

        Args:
            dest (str): The destination path where to place the converted value.
            src (str or None): The source from where the value to convert is fetched.
                When `src` is a path (ex. `context__user_id`), the value is the item
                    of the event at the path.
                When `src` is `None`, the value is the whole event.
            transformers (function or tuple of functions): The function(s) to apply on
                the source value.
            raw_input (bool): Flag indicating whether `get_value` will receive a raw
                event string or a parsed event dictionary.
        """
        object.__setattr__(self, "dest", tuple(dest.split(MODEL_PATH_SEPARATOR)))
        src = tuple(src.split(MODEL_PATH_SEPARATOR)) if src else None
        object.__setattr__(self, "src", src)
        transformers = (transformers,) if callable(transformers) else transformers
        object.__setattr__(self, "transformers", transformers)
        object.__setattr__(self, "raw_input", raw_input)

    def get_value(self, data: Union[Dict, str]) -> Union[Dict, str]:
        """Return fetched source value after having applied all transformers to it.

        Args:
            data (dict or string): The event to convert.

        Raises:
            ConversionException: When a field transformation fails.
        """
        if self.src:
            data = get_dict_value_from_path(data, self.src)
        try:
            for transformer in self.transformers:
                data = transformer(data)
        except Exception as err:  # noqa: BLE001
            msg = f"Failed to get the transformed value for field: {self.src}"
            raise ConversionException(msg) from err
        return data


class BaseConversionSet(ABC):
    """ConversionSet Base Class.

    ConversionSets are used to convert from one event format to another.
    """

    __src__: BaseModel
    __dest__: BaseModel

    def __init__(self) -> None:
        """Initialize BaseConversionSet."""
        self._conversion_items = self._get_conversion_items()

    @abstractmethod
    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""

    def __iter__(self) -> Iterator[ConversionItem]:  # noqa: D105
        return iter(self._conversion_items)


def convert_dict_event(
    event: dict, event_str: str, conversion_set: BaseConversionSet
) -> Any:
    """Convert the event dictionary with a conversion_set.

    Args:
        event (dict): The event to convert.
        event_str (dict): The original event string.
        conversion_set (BaseConversionSet): A conversion set used for conversion.

    Returns:
        event (BaseModel): The converted event pydantic model.

    Raises:
        ConversionException: When a field transformation fails.
        ValidationError: When the final converted event is invalid.
    """
    converted_event = {}
    for conversion_item in conversion_set:
        data = event_str if conversion_item.raw_input else event
        value = conversion_item.get_value(data)
        if value not in [None, "", {}]:
            set_dict_value_from_path(converted_event, conversion_item.dest, value)
    logger.debug("Intermediate converted event: %s", converted_event)
    return conversion_set.__dest__(**converted_event)


def convert_str_event(event_str: str, conversion_set: BaseConversionSet) -> BaseModel:
    """Convert the event string using the provided conversion_set.

    Args:
        event_str (str): The event to convert.
        conversion_set (BaseConversionSet): A conversion set used for conversion.

    Returns:
        event (BaseModel): The converted event pydantic model.

    Raises:
        BadFormatException: When the event_str is not a valid JSON string.
        ConversionException: When a field transformation fails.
        ValidationError: When the converted event is invalid.
    """
    try:
        event = json.loads(event_str)
    except (TypeError, json.JSONDecodeError) as err:
        msg = "Failed to parse the event, invalid JSON string"
        raise BadFormatException(msg) from err
    return convert_dict_event(event, event_str, conversion_set)


class Converter:
    """Events converter using pydantic models."""

    def __init__(
        self,
        model_selector: ModelSelector = ModelSelector(),
        module: str = "ralph.models.edx.converters.xapi",
        **conversion_set_kwargs: Any,
    ) -> None:
        """Initialize the Converter."""
        self.model_selector = model_selector
        self.src_conversion_set = self.get_src_conversion_set(
            import_module(module), **conversion_set_kwargs
        )

    @staticmethod
    def get_src_conversion_set(
        module: ModuleType, **conversion_set_kwargs: Any
    ) -> dict:
        """Return a dictionary of initialized conversion_sets defined in the module."""
        src_conversion_set = {}
        for _, class_ in getmembers(module, isclass):
            if issubclass(class_, BaseConversionSet):
                src_conversion_set[class_.__src__] = class_(**conversion_set_kwargs)
        return src_conversion_set

    def convert(
        self, input_file: TextIO, ignore_errors: bool, fail_on_unknown: bool
    ) -> Generator:
        """Convert JSON event strings line by line."""
        total = 0
        success = 0
        for event_str in input_file:
            try:
                total += 1
                yield self._convert_event(event_str).json(
                    exclude_none=True, by_alias=True
                )
                success += 1
            except (TypeError, json.JSONDecodeError) as err:
                message = "Input event is not a valid JSON string"
                self._log_error(message, event_str, err)
                if not ignore_errors:
                    raise BadFormatException(message) from err
            except (UnknownEventException, MissingConversionSetException) as err:
                self._log_error(err, event_str)
                if fail_on_unknown:
                    raise err
            except ConversionException as err:
                self._log_error(err, event_str)
                if not ignore_errors:
                    raise err
            except ValidationError as err:
                message = f"Converted event is not a valid ({err.model}) model"
                self._log_error(message, event_str, err)
                if not ignore_errors:
                    raise err
        logger.info("Total events: %d, Invalid events: %d", total, total - success)

    def _convert_event(self, event_str: str) -> Any:
        """Convert a single JSON string event.

        Args:
            event_str (str): The event to convert.

        Returns:
            event (BaseModel): The converted event pydantic model.

        Raises:
            TypeError: When the event_str is not of type string.
            JSONDecodeError: When the event_str is not a valid JSON string.
            UnknownEventException: When no matching model is found for the event.
            MissingConversionSetException: When no matching conversion set is found for
                the event.
            ConversionException: When a field transformation fails.
            ValidationError: When the final converted event is invalid.
        """
        error: Optional[BaseException] = None
        event = json.loads(event_str)
        for model in self.model_selector.get_models(event):
            conversion_set = self.src_conversion_set.get(model, None)
            if not conversion_set:
                message = "No conversion set found for input event"
                error = MissingConversionSetException(message)
                continue

            try:
                return convert_dict_event(event, event_str, conversion_set)
            except (ConversionException, ValidationError) as err:
                error = err

        raise error

    @staticmethod
    def _log_error(
        message: object, event_str: str, error: Optional[BaseException] = None
    ) -> None:
        logger.error(message)
        logger.debug("Raised error: %s, for event : %s", error, event_str)



"""S3 data backend for Ralph."""

import logging
from io import IOBase
from typing import Iterable, Iterator, Optional, Union
from uuid import uuid4

import boto3
from boto3.s3.transfer import TransferConfig
from botocore.client import BaseClient
from botocore.exceptions import (
    ClientError,
    EndpointConnectionError,
    ParamValidationError,
    ReadTimeoutError,
    ResponseStreamingError,
)
from pydantic import PositiveInt
from requests_toolbelt import StreamingIterator

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.backends.data.mixins import HistoryMixin
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now, parse_iterable_to_dict

logger = logging.getLogger(__name__)


class S3DataBackendSettings(BaseDataBackendSettings):
    """S3 data backend default configuration.

    Attributes:
        ACCESS_KEY_ID (str): The access key id for the S3 account.
        SECRET_ACCESS_KEY (str): The secret key for the S3 account.
        SESSION_TOKEN (str): The session token for the S3 account.
        ENDPOINT_URL (str): The endpoint URL of the S3.
        DEFAULT_REGION (str): The default region used in instantiating the client.
        DEFAULT_BUCKET_NAME (str): The default bucket name targeted.
        LOCALE_ENCODING (str): The encoding used for writing dictionaries to objects.
        READ_CHUNK_SIZE (str): The default chunk size for reading objects.
        WRITE_CHUNK_SIZE (str): The default chunk size for writing objects.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__S3__"

    ACCESS_KEY_ID: Optional[str] = None
    SECRET_ACCESS_KEY: Optional[str] = None
    SESSION_TOKEN: Optional[str] = None
    ENDPOINT_URL: Optional[str] = None
    DEFAULT_REGION: Optional[str] = None
    DEFAULT_BUCKET_NAME: Optional[str] = None
    READ_CHUNK_SIZE: int = 4096
    WRITE_CHUNK_SIZE: int = 4096


class S3DataBackend(
    BaseDataBackend[S3DataBackendSettings, str], Writable, Listable, HistoryMixin
):
    """S3 data backend."""

    name = "s3"
    default_operation_type = BaseOperationType.CREATE
    unsupported_operation_types = {
        BaseOperationType.APPEND,
        BaseOperationType.DELETE,
        BaseOperationType.UPDATE,
    }

    def __init__(self, settings: Optional[S3DataBackendSettings] = None):
        """Instantiate the AWS S3 client."""
        super().__init__(settings)
        self.default_bucket_name = self.settings.DEFAULT_BUCKET_NAME
        self._client = None

    @property
    def client(self) -> BaseClient:
        """Create a boto3 client if it doesn't exist."""
        if not self._client:
            self._client = boto3.client(
                "s3",
                aws_access_key_id=self.settings.ACCESS_KEY_ID,
                aws_secret_access_key=self.settings.SECRET_ACCESS_KEY,
                aws_session_token=self.settings.SESSION_TOKEN,
                region_name=self.settings.DEFAULT_REGION,
                endpoint_url=self.settings.ENDPOINT_URL,
            )
        return self._client

    def status(self) -> DataBackendStatus:
        """Implement data backend checks (e.g. connection, cluster status).

        Return:
            DataBackendStatus: The status of the data backend.
        """
        try:
            self.client.head_bucket(Bucket=self.default_bucket_name)
        except (ClientError, EndpointConnectionError):
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List objects for the target bucket.

        Args:
            target (str or None): The target bucket to list from.
                If target is `None`, the `default_bucket_name` is used instead.
            details (bool): Get detailed object information instead of just object name.
            new (bool): Given the history, list only unread files.

        Yield:
            str: The next object name. (If details is False).
            dict: The next object details. (If details is True).

        Raise:
            BackendException: If a failure occurs.
        """
        if target is None:
            target = self.default_bucket_name

        objects_to_skip = set()
        if new:
            objects_to_skip = set(self.get_command_history(self.name, "read"))

        try:
            paginator = self.client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=target)
            for objects in page_iterator:
                if "Contents" not in objects:
                    continue
                for obj in objects["Contents"]:
                    if new and f"{target}/{obj['Key']}" in objects_to_skip:
                        continue
                    if details:
                        obj["LastModified"] = obj["LastModified"].isoformat()
                        yield obj
                    else:
                        yield obj["Key"]
        except ClientError as err:
            error_msg = err.response["Error"]["Message"]
            msg = "Failed to list the bucket %s: %s"
            logger.error(msg, target, error_msg)
            raise BackendException(msg % (target, error_msg)) from err

    def read(  # noqa: PLR0913
        self,
        query: Optional[str] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read an object matching the `query` in the `target` bucket and yield it.

        Args:
            query (str): The ID of the object to read.
            target (str or None): The target bucket containing the object.
                If target is `None`, the `default_bucket` is used instead.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a connection failure occurs while reading from S3 or
                during object encoding and `ignore_errors` is set to `False`.
            BackendParameterException: If a backend argument value is not valid.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_bytes(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        target = self.default_bucket_name if target is None else target
        response = self._get_object(target, query)
        try:
            yield from response["Body"].iter_chunks(chunk_size)
        except (ReadTimeoutError, ResponseStreamingError) as err:
            msg = "Failed to read chunk from object %s"
            logger.error(msg, query)
            raise BackendException(msg % (query)) from err

        # Archive fetched, add a new entry to the history.
        self.append_to_history(
            {
                "backend": self.name,
                "action": "read",
                "id": target + "/" + query,
                "size": response["ContentLength"],
                "timestamp": now(),
            }
        )

    def _read_dicts(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        target = self.default_bucket_name if target is None else target
        response = self._get_object(target, query)
        try:
            lines = response["Body"].iter_lines(chunk_size)
            yield from parse_iterable_to_dict(lines, ignore_errors)
        except (ReadTimeoutError, ResponseStreamingError) as err:
            msg = "Failed to read chunk from object %s"
            logger.error(msg, query)
            raise BackendException(msg % query) from err

        # Archive fetched, add a new entry to the history.
        self.append_to_history(
            {
                "backend": self.name,
                "action": "read",
                "id": target + "/" + query,
                "size": response["ContentLength"],
                "timestamp": now(),
            }
        )

    def _get_object(self, bucket: Optional[str], key: str) -> dict:
        """Validate bucket (target) and key (query) and return the S3 object."""
        if not bucket:
            msg = "The target bucket is not set"
            logger.error(msg)
            raise BackendParameterException(msg)

        if not key:
            msg = "The object ID query is not set"
            logger.error(msg)
            raise BackendParameterException(msg)

        try:
            return self.client.get_object(Bucket=bucket, Key=key)
        except (ClientError, EndpointConnectionError) as error:
            error_msg = error
            if isinstance(error, ClientError):
                error_msg = error.response["Error"]["Message"]

            msg = "Failed to download %s: %s"
            logger.error(msg, key, error_msg)
            raise BackendException(msg % (key, error_msg)) from error

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write `data` records to the `target` bucket and return their count.

        Args:
            data (Iterable or IOBase): The data to write.
            target (str or None): The target bucket and the target object
                separated by a `/`.
                If target is `None`, the default bucket is used and a random
                (uuid4) object is created.
                If target does not contain a `/`, it is assumed to be the
                target object and the default bucket is used.
            chunk_size (int or None): The chunk size when writing objects by batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding and encoding of
                records are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write
                operation.
                If operation_type is `CREATE` or `INDEX`, the target object is
                expected to be absent. If the target object exists a
                `BackendException` is raised.

        Return:
            int: The number of written objects.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If a backend argument value is not valid.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        return super()._write_dicts(
            data, target, chunk_size, ignore_errors, operation_type
        )

    def _write_bytes(  # noqa: PLR0913
        self,
        data: Iterable[bytes],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing bytes. See `self.write`."""
        if not target:
            target = f"{self.default_bucket_name}/{now()}-{uuid4()}"
            logger.info(
                "Target not specified; using default bucket with random file name: %s",
                target,
            )
        elif "/" not in target:
            target = f"{self.default_bucket_name}/{target}"
            logger.info(
                "Target not specified; using default bucket: %s",
                target,
            )

        target_bucket, target_object = target.split("/", 1)
        if target_object in list(self.list(target=target_bucket)):
            msg = "%s already exists and overwrite is not allowed for operation %s"
            logger.error(msg, target_object, operation_type)
            raise BackendException(msg % (target_object, operation_type))

        logger.info("Creating archive: %s", target_object)
        counter = {"count": 0}
        data = self._count(data, counter)

        # Using StreamingIterator from requests-toolbelt but without specifying a size
        # as we will not use it. It implements the `read` method for iterators.
        file_object = StreamingIterator(0, data)

        try:
            self.client.upload_fileobj(
                Bucket=target_bucket,
                Key=target_object,
                Fileobj=file_object,
                Config=TransferConfig(multipart_chunksize=chunk_size),
            )
            response = self.client.head_object(Bucket=target_bucket, Key=target_object)
        except (ClientError, ParamValidationError, EndpointConnectionError) as error:
            msg = "Failed to upload %s"
            logger.error(msg, target)
            raise BackendException(msg % target) from error

        # Archive written, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "action": "write",
                "operation_type": operation_type.value,
                "id": target,
                "size": response["ContentLength"],
                "timestamp": now(),
            }
        )

        return counter["count"]

    def close(self) -> None:
        """Close the S3 backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        try:
            self.client.close()
        except ClientError as error:
            msg = "Failed to close S3 backend client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    @staticmethod
    def _count(statements: Iterable, counter: dict) -> Iterator:
        """Count the elements in the `statements` Iterable and yield element."""
        for statement in statements:
            yield statement
            counter["count"] += 1



"""Elasticsearch data backend for Ralph."""

import logging
from io import IOBase
from pathlib import Path
from typing import Iterable, Iterator, List, Literal, Optional, TypeVar, Union

from elasticsearch import ApiError, Elasticsearch, TransportError
from elasticsearch.helpers import BulkIndexError, streaming_bulk
from pydantic import BaseModel, PositiveInt, ValidationError
from typing_extensions import Self

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    BaseQuery,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.conf import BaseSettingsConfig, ClientOptions, CommaSeparatedTuple
from ralph.exceptions import BackendException

logger = logging.getLogger(__name__)


class ESClientOptions(ClientOptions):
    """Elasticsearch additional client options."""

    ca_certs: Optional[Path] = None
    verify_certs: Optional[bool] = None


class ESDataBackendSettings(BaseDataBackendSettings):
    """Elasticsearch data backend default configuration.

    Attributes:
        ALLOW_YELLOW_STATUS (bool): Whether to consider Elasticsearch yellow health
            status to be ok.
        CLIENT_OPTIONS (dict): A dictionary of valid options for the Elasticsearch class
            initialization.
        DEFAULT_INDEX (str): The default index to use for querying Elasticsearch.
        HOSTS (str or tuple): The comma-separated list of Elasticsearch nodes to
            connect to.
        LOCALE_ENCODING (str): The encoding used for reading/writing documents.
        POINT_IN_TIME_KEEP_ALIVE (str): The duration for which Elasticsearch should
            keep a point in time alive.
        READ_CHUNK_SIZE (int): The default chunk size for reading batches of documents.
        REFRESH_AFTER_WRITE (str or bool): Whether the Elasticsearch index should be
            refreshed after the write operation.
        WRITE_CHUNK_SIZE (int): The default chunk size for writing batches of documents.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__ES__"

    ALLOW_YELLOW_STATUS: bool = False
    CLIENT_OPTIONS: ESClientOptions = ESClientOptions()
    DEFAULT_INDEX: str = "statements"
    HOSTS: CommaSeparatedTuple = ("http://localhost:9200",)
    POINT_IN_TIME_KEEP_ALIVE: str = "1m"
    REFRESH_AFTER_WRITE: Union[Literal["false", "true", "wait_for"], bool, str, None]


class ESQueryPit(BaseModel):
    """Elasticsearch point in time (pit) query configuration.

    Attributes:
        id (str): Context identifier of the Elasticsearch point in time.
        keep_alive (str): The duration for which Elasticsearch should keep the point in
            time alive.
    """

    id: Union[str, None]
    keep_alive: Union[str, None]


class ESQuery(BaseQuery):
    """Elasticsearch query model.

    Attributes:
        q (str): The Elastisearch query in the Lucene query string syntax.
            See Elasticsearch search reference for Lucene query syntax:
            https://www.elastic.co/guide/en/elasticsearch/reference/8.9/search-search.html#search-api-query-params-q
        query (dict): A search query definition using the Elasticsearch Query DSL.
            See Elasticsearch search reference for query DSL syntax:
            https://www.elastic.co/guide/en/elasticsearch/reference/8.9/search-search.html#request-body-search-query
        pit (dict): Limit the search to a point in time (PIT). See ESQueryPit.
        size (int): The maximum number of documents to yield.
        sort (str or list): Specify how to sort search results. Set to `_doc` or
            `_shard_doc` if order doesn't matter.
            See https://www.elastic.co/guide/en/elasticsearch/reference/8.9/sort-search-results.html
        search_after (list): Limit search query results to values after a document
            matching the set of sort values in `search_after`. Used for pagination.
        track_total_hits (bool): Number of hits matching the query to count accurately.
            Not used. Always set to `False`.
    """

    q: Optional[str] = None
    query: dict = {"match_all": {}}
    pit: ESQueryPit = ESQueryPit()
    size: Optional[int] = None
    sort: Union[str, List[dict]] = "_shard_doc"
    search_after: Optional[list] = None
    track_total_hits: Literal[False] = False

    @classmethod
    def from_string(cls, query: str) -> Self:
        """Return an instance of itself from a string."""
        try:
            return cls.parse_raw(query)
        except ValidationError:
            msg = "Fallback to Lucene Query as the query is not an ESQuery: %s"
            logger.info(msg, query)
            return cls(q=query)


Settings = TypeVar("Settings", bound=ESDataBackendSettings)


class ESDataBackend(BaseDataBackend[Settings, ESQuery], Writable, Listable):
    """Elasticsearch data backend."""

    name = "es"
    unsupported_operation_types = {BaseOperationType.APPEND}

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the Elasticsearch data backend.

        Args:
            settings (Settings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self._client = None

    @property
    def client(self) -> Elasticsearch:
        """Create an Elasticsearch client if it doesn't exist."""
        if not self._client:
            self._client = Elasticsearch(
                self.settings.HOSTS, **self.settings.CLIENT_OPTIONS.dict()
            )
        return self._client

    def status(self) -> DataBackendStatus:
        """Check Elasticsearch cluster connection and status."""
        try:
            self.client.info()
            cluster_status = self.client.cat.health()
        except TransportError as error:
            logger.error("Failed to connect to Elasticsearch: %s", error)
            return DataBackendStatus.AWAY

        if "green" in cluster_status:
            return DataBackendStatus.OK

        if "yellow" in cluster_status and self.settings.ALLOW_YELLOW_STATUS:
            logger.info("Cluster status is yellow.")
            return DataBackendStatus.OK

        logger.error("Cluster status is not green: %s", cluster_status)

        return DataBackendStatus.ERROR

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List available Elasticsearch indices, data streams and aliases.

        Args:
            target (str or None): The comma-separated list of data streams, indices,
                and aliases to limit the request. Supports wildcards (*).
                If target is `None`, lists all available indices, data streams and
                    aliases. Equivalent to (`target` = "*").
            details (bool): Get detailed information instead of just names.
            new (bool): Ignored.

        Yield:
            str: The next index, data stream or alias name. (If `details` is False).
            dict: The next index, data stream or alias details. (If `details` is True).

        Raise:
            BackendException: If a failure during indices retrieval occurs.
        """
        target = target if target else "*"
        try:
            indices = self.client.indices.get(index=target)
        except (ApiError, TransportError) as error:
            msg = "Failed to read indices: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        if new:
            logger.warning("The `new` argument is ignored")

        if details:
            for index, value in indices.items():
                yield {index: value}

            return

        for index in indices:
            yield index

    def read(  # noqa: PLR0913
        self,
        query: Optional[ESQuery] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read documents matching the query in the target index and yield them.

        Args:
            query (ESQuery): The Elasticsearch query to use when fetching documents.
            target (str or None): The target Elasticsearch index name to query.
                If target is `None`, the `DEFAULT_INDEX` is used instead.
            chunk_size (int or None): The chunk size when reading documents by batches.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield dictionaries or bytes.
            ignore_errors (bool): No impact as encoding errors are not expected in
                Elasticsearch results.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The next raw document if `raw_output` is True.
            dict: The next JSON parsed document if `raw_output` is False.

        Raise:
            BackendException: If a failure occurs during Elasticsearch connection.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_dicts(
        self,
        query: ESQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        target = target if target else self.settings.DEFAULT_INDEX
        if not query.pit.keep_alive:
            query.pit.keep_alive = self.settings.POINT_IN_TIME_KEEP_ALIVE
        if not query.pit.id:
            try:
                query.pit.id = self.client.open_point_in_time(
                    index=target, keep_alive=query.pit.keep_alive
                )["id"]
            except (ApiError, TransportError, ValueError) as error:
                msg = "Failed to open Elasticsearch point in time: %s"
                logger.error(msg, error)
                raise BackendException(msg % error) from error

        limit = query.size
        kwargs = query.dict()
        count = chunk_size
        # The first condition is set to comprise either limit as None
        # (when the backend query does not have `size` parameter),
        # or limit with a positive value.
        while limit != 0 and chunk_size == count:
            kwargs["size"] = limit if limit and limit < chunk_size else chunk_size
            try:
                documents = self.client.search(**kwargs)["hits"]["hits"]
            except (ApiError, TransportError, TypeError) as error:
                msg = "Failed to execute Elasticsearch query: %s"
                logger.error(msg, error)
                raise BackendException(msg % error) from error
            count = len(documents)
            if limit:
                limit -= count if chunk_size == count else limit
            query.search_after = None
            if count:
                query.search_after = [str(part) for part in documents[-1]["sort"]]
            kwargs["search_after"] = query.search_after
            yield from documents

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write data documents to the target index and return their count.

        Args:
            data (Iterable or IOBase): The data containing documents to write.
            target (str or None): The target Elasticsearch index name.
                If target is `None`, the `DEFAULT_INDEX` is used instead.
            chunk_size (int or None): The number of documents to write in one batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding, encoding and
                sending batches of documents are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.

        Return:
            int: The number of documents written.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If the `operation_type` is `APPEND` as it is not
                supported.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        count = 0
        target = target if target else self.settings.DEFAULT_INDEX
        msg = "Start writing to the %s index (chunk size: %d)"
        logger.debug(msg, target, chunk_size)
        try:
            for success, action in streaming_bulk(
                client=self.client,
                actions=ESDataBackend.to_documents(data, target, operation_type),
                chunk_size=chunk_size,
                raise_on_error=(not ignore_errors),
                refresh=self.settings.REFRESH_AFTER_WRITE,
            ):
                count += success
                logger.debug("Wrote %d document [action: %s]", success, action)

            logger.info("Finished writing %d documents with success", count)
        except (BulkIndexError, ApiError, TransportError) as error:
            msg = "%s %s Total succeeded writes: %s"
            details = getattr(error, "errors", "")
            logger.error(msg, error, details, count)
            raise BackendException(msg % (error, details, count)) from error
        return count

    def close(self) -> None:
        """Close the Elasticsearch backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        try:
            self.client.close()
        except TransportError as error:
            msg = "Failed to close Elasticsearch client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    @staticmethod
    def to_documents(
        data: Iterable[dict],
        target: str,
        operation_type: BaseOperationType,
    ) -> Iterator[dict]:
        """Convert dictionaries from `data` to ES documents and yield them."""
        if operation_type == BaseOperationType.UPDATE:
            for item in data:
                yield {
                    "_index": target,
                    "_id": item.get("id", None),
                    "_op_type": operation_type.value,
                    "doc": item,
                }
        elif operation_type in (BaseOperationType.CREATE, BaseOperationType.INDEX):
            for item in data:
                yield {
                    "_index": target,
                    "_id": item.get("id", None),
                    "_op_type": operation_type.value,
                    "_source": item,
                }
        else:
            # operation_type == BaseOperationType.DELETE (by exclusion)
            for item in data:
                yield {
                    "_index": target,
                    "_id": item.get("id", None),
                    "_op_type": operation_type.value,
                }



"""Ralph backend loader."""

import logging
import pkgutil
from functools import lru_cache
from importlib import import_module
from importlib.util import find_spec
from inspect import getmembers, isabstract, isclass
from typing import Dict, Tuple, Type

from ralph.backends.data.base import (
    AsyncListable,
    AsyncWritable,
    BaseAsyncDataBackend,
    BaseDataBackend,
    Listable,
    Writable,
)
from ralph.backends.lrs.base import BaseAsyncLRSBackend, BaseLRSBackend

logger = logging.getLogger(__name__)


@lru_cache()
def get_backends(packages: Tuple[str], base_backends: Tuple[Type]) -> Dict[str, Type]:
    """Return sub-classes of `base_backends` found in sub-modules of `packages`.

    Args:
        packages (tuple of str): A tuple of dotted package names.
            Ex.: ("ralph.backends.data", "ralph.backends.lrs").
        base_backends (tuple of type): A tuple of base backend classes to search for.
            Ex.: ("BaseDataBackend",)

    Return:
        dict: A dictionary of found non-abstract backend classes by their name property.
            Ex.: {"fs": FSDataBackend}
    """
    module_specs = []
    for package in packages:
        try:
            module_spec = find_spec(package)
        except ModuleNotFoundError:
            module_spec = None

        if not module_spec:
            logger.warning("Could not find '%s' package; skipping it", package)
            continue

        module_specs.append(module_spec)

    modules = []
    for module_spec in module_specs:
        paths = module_spec.submodule_search_locations
        for module_info in pkgutil.iter_modules(paths, prefix=f"{module_spec.name}."):
            modules.append(module_info.name)

    backend_classes = set()
    for module in modules:
        try:
            backend_module = import_module(module)
        except Exception as error:  # noqa: BLE001
            logger.debug("Failed to import %s module: %s", module, error)
            continue
        for _, class_ in getmembers(backend_module, isclass):
            if issubclass(class_, base_backends) and not isabstract(class_):
                backend_classes.add(class_)

    return {
        backend.name: backend
        for backend in sorted(backend_classes, key=lambda x: x.name, reverse=True)
    }


@lru_cache(maxsize=1)
def get_cli_backends() -> Dict[str, Type]:
    """Return Ralph's backend classes for cli usage."""
    base_backends = (BaseAsyncDataBackend, BaseDataBackend)
    return get_backends(("ralph.backends.data",), base_backends)


@lru_cache(maxsize=1)
def get_cli_write_backends() -> Dict[str, Type]:
    """Return Ralph's backends classes for cli write usage."""
    return {
        name: backend
        for name, backend in get_cli_backends().items()
        if issubclass(backend, (Writable, AsyncWritable))
    }


@lru_cache(maxsize=1)
def get_cli_list_backends() -> Dict[str, Type]:
    """Return Ralph's backends classes for cli list usage."""
    return {
        name: backend
        for name, backend in get_cli_backends().items()
        if issubclass(backend, (Listable, AsyncListable))
    }


@lru_cache(maxsize=1)
def get_lrs_backends() -> Dict[str, Type]:
    """Return Ralph's backend classes for LRS usage."""
    return get_backends(("ralph.backends.lrs",), (BaseAsyncLRSBackend, BaseLRSBackend))
